import torch
from antgo.framework.helper.parallel import MMDataParallel, MMDistributedDataParallel

dp_factory = {'cuda': MMDataParallel, 'cpu': MMDataParallel}
ddp_factory = {'cuda': MMDistributedDataParallel}


def build_dp(model, device='cuda', dim=0, *args, **kwargs):
    """build DataParallel module by device type.

    if device is cuda, return a MMDataParallel model; if device is mlu,
    return a MLUDataParallel model.

    Args:
        model (:class:`nn.Module`): model to be parallelized.
        device (str): device type, cuda, cpu or mlu. Defaults to cuda.
        dim (int): Dimension used to scatter the data. Defaults to 0.

    Returns:
        nn.Module: the model to be parallelized.
    """
    if device == 'cuda':
        model = model.cuda()

    return dp_factory[device](model, dim=dim, *args, **kwargs)


def build_ddp(model, device='cuda', *args, **kwargs):
    """Build DistributedDataParallel module by device type.

    If device is cuda, return a MMDistributedDataParallel model;
    if device is mlu, return a MLUDistributedDataParallel model.

    Args:
        model (:class:`nn.Module`): module to be parallelized.
        device (str): device type, mlu or cuda.

    Returns:
        :class:`nn.Module`: the module to be parallelized

    References:
        .. [1] https://pytorch.org/docs/stable/generated/torch.nn.parallel.
                     DistributedDataParallel.html
    """
    assert device in ['cuda', 'mlu'], 'Only available for cuda or mlu devices.'
    if device == 'cuda':
        model = model.cuda()

    return ddp_factory[device](model, *args, **kwargs)


def is_mlu_available():
    """Returns a bool indicating if MLU is currently available."""
    return hasattr(torch, 'is_mlu_available') and torch.is_mlu_available()


def get_device():
    """Returns an available device, cpu, cuda or mlu."""
    is_device_available = {
        'cuda': torch.cuda.is_available(),
        'mlu': is_mlu_available()
    }
    device_list = [k for k, v in is_device_available.items() if v]
    return device_list[0] if len(device_list) == 1 else 'cpu'



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from ...builder import DISTILL_LOSSES



'''
CC with P-order Taylor Expansion of Gaussian RBF kernel
'''

@DISTILL_LOSSES.register_module()
class CC(nn.Module):
	'''
	Correlation Congruence for Knowledge Distillation
	http://openaccess.thecvf.com/content_ICCV_2019/papers/
	Peng_Correlation_Congruence_for_Knowledge_Distillation_ICCV_2019_paper.pdf
	'''
	def __init__(self, gamma, P_order):
		super(CC, self).__init__()
		self.gamma = gamma
		self.P_order = P_order

	def forward(self, feat_s, feat_t):
		corr_mat_s = self.get_correlation_matrix(feat_s)
		corr_mat_t = self.get_correlation_matrix(feat_t)

		loss = F.mse_loss(corr_mat_s, corr_mat_t)

		return loss

	def get_correlation_matrix(self, feat):
		feat = F.normalize(feat, p=2, dim=-1)
		sim_mat  = torch.matmul(feat, feat.t())
		corr_mat = torch.zeros_like(sim_mat)

		for p in range(self.P_order+1):
			corr_mat += math.exp(-2*self.gamma) * (2*self.gamma)**p / \
						math.factorial(p) * torch.pow(sim_mat, p)

		return corr_mat


import enum
from antgo.framework.helper.models.builder import DETECTORS
from ..single_stage import SingleStageDetector, to_image_list
import torch
from torch import nn
import torch.nn.functional as F


@DETECTORS.register_module()
class TTFNet(SingleStageDetector):
    def __init__(self,
                 backbone,
                 neck,
                 bbox_head,
                 train_cfg=None,
                 test_cfg=None,
                 pretrained=None,
                 init_cfg=None):
        super(TTFNet, self).__init__(backbone, neck, bbox_head,
                                     train_cfg, test_cfg, pretrained, init_cfg)

    def simple_test(self, image, image_meta, rescale=True, **kwargs):
        """Test function without test-time augmentation.

        Args:
            img (torch.Tensor): Images with shape (N, C, H, W).
            img_metas (list[dict]): List of image information.
            rescale (bool, optional): Whether to rescale the results.
                Defaults to False.

        Returns:
            list[list[np.ndarray]]: BBox results of each image and classes.
                The outer list corresponds to each image. The inner list
                corresponds to each class.
        """
        image_list, image_meta = to_image_list(image, image_meta)
        image = image_list.tensors
        
        feat = self.extract_feat(image)
        results_list = self.bbox_head.simple_test(
            feat, image_meta, rescale=rescale)

        bbox_results = {
            'box': [a for a, _ in results_list],
            'label': [b for _, b in results_list],
        }
        # {'box', 'label'}
        return bbox_results

    def onnx_export(self, image):
        feat = self.extract_feat(image)
        local_cls, local_reg = self.bbox_head(feat)
        
        if isinstance(local_cls, list):
            for level_i, level_local_cls in enumerate(local_cls):
                level_local_cls = torch.sigmoid(level_local_cls)
                level_local_cls = F.max_pool2d(level_local_cls, 3, stride=1, padding=(3 - 1) // 2)
                local_cls[level_i] = level_local_cls
            
            return local_cls, local_reg
        else:
            local_cls = torch.sigmoid(local_cls)
            local_cls = F.max_pool2d(local_cls, 3, stride=1, padding=(3 - 1) // 2)
            return local_cls, local_reg



# -*- coding: UTF-8 -*-
# @Time    : 2022/5/2 12:28
# @File    : io.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import torch
import torch


class DataLoader(torch.utils.data.DataLoader):
    def __init__(self,
                 dataset,
                 feed_list=None,
                 places=None,
                 return_list=True,
                 batch_sampler=None,
                 batch_size=1,
                 shuffle=False,
                 drop_last=False,
                 collate_fn=None,
                 num_workers=0,
                 use_buffer_reader=True,
                 use_shared_memory=True,
                 timeout=0,
                 worker_init_fn=None):
        assert(return_list)
        super().__init__(dataset,
                         batch_size=batch_size,
                         shuffle=shuffle,
                         sampler=None,
                         batch_sampler=batch_sampler,
                         num_workers=num_workers,
                         collate_fn=collate_fn,
                         pin_memory=False,
                         drop_last=drop_last,
                         timeout=timeout,
                         worker_init_fn=worker_init_fn,
                         multiprocessing_context=None,
                         prefetch_factor=2,
                         persistent_workers=False)




# encoding=utf-8
# @Time    : 17-7-12
# @File    : segmentation_task.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import sys
import numpy as np
from antgo.task.task import *
from antgo.measures.base import *
from antgo.dataflow.common import *
from scipy import signal
from scipy import ndimage
from scipy.sparse import csr_matrix


default = {'AntMeanIOUSeg': ('MeanIOU', 'SEGMENTATION'),
           'AntMeanIOUBoundary': ('MeanIOUBoundary', 'SEGMENTATION'),
           }

def get_confusion_matrix(gt_label, pred_label, num_classes):
    """
    Calcute the confusion matrix by given label and pred
    :param gt_label: the ground truth label
    :param pred_label: the pred label
    :param num_classes: the nunber of class
    :return: the confusion matrix
    """
    index = (gt_label * num_classes + pred_label).astype('int32')
    label_count = np.bincount(index)
    confusion_matrix = np.zeros((num_classes, num_classes))

    for i_label in range(num_classes):
        for i_pred_label in range(num_classes):
            cur_index = i_label * num_classes + i_pred_label
            if cur_index < len(label_count):
                confusion_matrix[i_label, i_pred_label] = label_count[cur_index]

    return confusion_matrix


class AntMeanIOUSeg(AntMeasure):
  def __init__(self, task):
    # paper: Jonathan Long, Evan Shelhamer, etc. Fully Convolutional Networks for Semantic Segmentation
    # formular: (1/n_{cl}) / \sum_i n_{ii}/(t_i+\sum_j n_{ji}-n_{ii})

    super(AntMeanIOUSeg, self).__init__(task, 'MeanIOU')
    assert(task.task_type == 'SEGMENTATION')
    self.classes_num = len(self.task.class_label)
    self.is_support_rank = True

  def _fast_hist(self, label_pred, label_true):
    # 找出标签中需要计算的类别,去掉了背景
    mask = (label_true >= 0) & (label_true < self.classes_num)

    # # np.bincount计算了从0到n**2-1这n**2个数中每个数出现的次数，返回值形状(n, n)
    hist = np.bincount(
      self.classes_num * label_true[mask].astype(np.int) +
      label_pred[mask].astype(np.int), minlength=self.classes_num ** 2).reshape(self.classes_num, self.classes_num)
    return hist

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    sample_scores = []
    confusion = np.zeros((self.classes_num, self.classes_num))
    for predict, gt in data:
      id = None
      if type(gt) == dict:
        if 'segmentation_map' not in gt:
          continue

        id = gt['id']
        gt = gt['segmentation_map']

      sample_confusion = self._fast_hist(np.array(predict).flatten(), np.array(gt).flatten())
      confusion += sample_confusion

      if id is not None:
        sample_iou = np.diag(sample_confusion) / np.maximum(1.0, sample_confusion.sum(axis=1) + sample_confusion.sum(axis=0) - np.diag(sample_confusion))
        sample_miou_val = np.nanmean(sample_iou)
        sample_scores.append({'id': id, 'score': sample_miou_val, 'category': 0})

    iou = np.diag(confusion) / np.maximum(1.0,confusion.sum(axis=1) + confusion.sum(axis=0) - np.diag(confusion))
    miou_val = float(np.nanmean(iou))

    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': miou_val, 'type':'SCALAR'}]},
            'info': sample_scores}


class AntMeanIOUBoundary(AntMeasure):
  def __init__(self, task):
    # paper: Liang-Chieh Chen, etc. Semantic Image Segmentation with Deep Convolutional Nets and Fully Connected CRFs
    # formular: (1/n_{cl}) / \sum_i n_{ii}/(t_i+\sum_j n_{ji}-n_{ii})

    super(AntMeanIOUBoundary, self).__init__(task, 'MeanIOUBoundary')
    assert(task.task_type == 'SEGMENTATION')
    self.is_support_rank = True
    self.classes_num = len(self.task.class_label)

  def _fast_hist(self, label_pred, label_true):
    # 找出标签中需要计算的类别,去掉了背景
    mask = (label_true >= 0) & (label_true < self.classes_num)

    # # np.bincount计算了从0到n**2-1这n**2个数中每个数出现的次数，返回值形状(n, n)
    hist = np.bincount(
      self.classes_num * label_true[mask].astype(int) +
      label_pred[mask].astype(int), minlength=self.classes_num ** 2).reshape(self.classes_num, self.classes_num)
    return hist

  def eva(self, data, label):
    trimap_width = int(getattr(self.task, 'MeanIOUBoundary_trimap_width', 3))
    if label is not None:
        data = zip(data, label)

    sample_scores = []
    confusion = np.zeros((self.classes_num, self.classes_num))
    for predict, gt in data:
      id = None
      if type(gt) == dict:
        if 'segmentation_map' not in gt:
          continue

        id = gt['id']
        gt = gt['segmentation_map']

      # finding object edge, ignore class 0
      ignore_gt = np.array(gt).astype(np.int)

      # ignore non edge region
      for l in range(1, self.classes_num):
        # finding edge region
        _gt = np.zeros(gt.shape)
        _gt[np.where(ignore_gt == l)] = 1.0

        _gt_blur = ndimage.gaussian_filter(_gt, sigma=7)
        _gt_blur[np.where(_gt_blur > 0.99)] = 0.0
        _gt_blur[np.where(_gt_blur < 0.01)] = 0.0
        _gt_blur[np.where(_gt_blur > 0.0)] = 1.0

        # ignore non edge region
        ignore_gt[np.where(_gt_blur < 1.0)] = 255

      sample_confusion = self._fast_hist(np.array(predict).flatten(), ignore_gt.flatten())
      confusion += sample_confusion

      if id is not None:
        sample_iou = np.diag(sample_confusion) / np.maximum(1.0, sample_confusion.sum(axis=1) + sample_confusion.sum(axis=0) - np.diag(sample_confusion))
        sample_miou_val = float(np.nanmean(sample_iou))
        sample_scores.append({'id': id, 'score': sample_miou_val, 'category': 0})

    iou = np.diag(confusion) / np.maximum(1.0, confusion.sum(axis=1) + confusion.sum(axis=0) - np.diag(confusion))
    miou_val = float(np.nanmean(iou))
    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': miou_val, 'type':'SCALAR'}]},
            'info': sample_scores}


from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
from typing import List
from functools import partial
import pickle
import os
import tfrecord
from tfrecord.tools.tfrecord2idx import *
from antgo.ant import environment
import json
import logging


numpy_dtype_map = {
    0: np.float32,
    1: np.int32,
    2: np.int64,
    3: np.int8,
    4: np.uint8
}

class TFDataWriter(object):
    def __init__(self, prefix, output_path, size_in_shard=100000, num_shards=1, keys=[]) -> None:
        self.num_shards = num_shards
        self.prefix = prefix
        self.output_path = output_path
        self.size_in_shard = size_in_shard

    def write(self, data_iterator):
        data_num = len(data_iterator)     
        size_in_shard = self.size_in_shard
        if size_in_shard < 0:
            size_in_shard = (data_num+self.num_shards-1) // self.num_shards
        else:
            num_shards = data_num // size_in_shard
            if num_shards * size_in_shard != data_num:
                num_shards += 1
            self.num_shards = num_shards

        tfwriter = None
        shard_i = -1
        count = 0

        # step 1: write tfrecord
        for index, data in enumerate(data_iterator):
            if data is None:
                continue

            data_and_description = {}
            for k,v in data.items():
                if isinstance(v, str):
                    data_and_description[k] = (v.encode('utf-8'), 'byte')
                elif isinstance(v, bytes):
                    data_and_description[k] = (v, 'byte')
                elif isinstance(v, np.ndarray):
                    if v.dtype not in [np.float32, np.int32, np.int64, np.int8, np.uint8]:
                        continue

                    data_and_description[k] = (v.tobytes(), 'byte')
                    data_and_description[f'__{k}_shape'] = (list(v.shape), 'int')
                    if v.dtype == np.float32:
                        data_and_description[f'__{k}_type'] = (0, 'int')
                    elif v.dtype == np.int32:
                        data_and_description[f'__{k}_type'] = (1, 'int')
                    elif v.dtype == np.int64:
                        data_and_description[f'__{k}_type'] = (2, 'int')
                    elif v.dtype == np.int8:
                        data_and_description[f'__{k}_type'] = (3, 'int')
                    elif v.dtype == np.uint8:
                        data_and_description[f'__{k}_type'] = (4, 'int')
                elif isinstance(v, list):
                    if isinstance(v[0], float):
                        data_and_description[k] = (v, 'float')
                    elif isinstance(v[0], int):
                        data_and_description[k] = (v, 'int')
                elif isinstance(v, float):
                    data_and_description[k] = (v, 'float')
                elif isinstance(v, int):
                    data_and_description[k] = (v, 'int')
                elif isinstance(v, dict):
                    data_and_description[k] = (json.dumps(v).encode('utf-8'), 'byte')

                if k not in data_and_description:
                    logging.error(f'ignore {k} in data')

            if len(data_and_description) == 0:
                continue

            if (count // size_in_shard) != shard_i:
                if tfwriter is not None:
                    tfwriter.close()
                shard_i = count // size_in_shard
                filename = os.path.join(self.output_path, "%s-%.5d-of-%.5d-tfrecord"%(self.prefix, shard_i, self.num_shards))
                tfwriter = tfrecord.TFRecordWriter(filename)

            if (index+1) % data_num == 0:
                logging.info(f'Finish tfrecord package process {index+1}/{data_num}.')

            tfwriter.write(data_and_description)            
            count += 1

        if tfwriter is not None:
            tfwriter.close()

        # step 2: create index
        for shard_i in range(self.num_shards):
            tfrecord_file = os.path.join(self.output_path, "%s-%.5d-of-%.5d-tfrecord"%(self.prefix, shard_i, self.num_shards))
            index_file = os.path.join(self.output_path, "%s-%.5d-of-%.5d-index"%(self.prefix, shard_i, self.num_shards))
            create_index(tfrecord_file, index_file)

        logging.info(f'Finish tfrecord index.')


class KVDataWriter(object):
    def __init__(self, prefix, output_path, size_in_shard=-1, num_shards=1, keys=[]) -> None:
        self.num_shards = num_shards
        self.prefix = prefix
        self.output_path = output_path
        self.size_in_shard = size_in_shard
        self.keys = keys

    def write(self, data_iterator):
        data_num = len(data_iterator)     
        size_in_shard = self.size_in_shard
        if size_in_shard < 0:
            size_in_shard = (data_num+self.num_shards-1) // self.num_shards
        else:
            num_shards = data_num // size_in_shard
            if num_shards * size_in_shard != data_num:
                num_shards += 1
            self.num_shards = num_shards

        kvwriter = None
        shard_i = -1

        cache_list = []
        cache_size = 1000
        for index, data in enumerate(data_iterator):
            if data is None:
                continue
            serial_data_bytes = pickle.dumps(data)

            if (index // size_in_shard) != shard_i:
                if kvwriter is not None:
                    if len(cache_list) > 0:
                        keys, values = zip(*cache_list)
                        kvwriter.write_many(keys, values)
                        cache_list = []

                    kvwriter.flush()

                # 构建新的数据包写对象
                shard_i = index // size_in_shard
                filename = os.path.join(self.output_path, "%s-%.5d-of-%.5d-kvrecord"%(self.prefix, shard_i, self.num_shards))
                kvwriter = environment.KVWriter(filename, self.num_shards)

            suffix = ''
            for k in self.keys:
                suffix += f'-{data[k]}'

            cache_list.append((f'{self.prefix}-{index}{suffix}', serial_data_bytes))
            if len(cache_list) == cache_size:
                logging.info(f'Finish kv package process {index+1}/{data_num}.')
                keys, values = zip(*cache_list)
                kvwriter.write_many(keys, values)
                cache_list = []

            index += 1

        if len(cache_list) > 0:
            keys, values = zip(*cache_list)
            kvwriter.write_many(keys, values)

        kvwriter.flush()
        logging.info(f'Finish kv package {data_num}/{data_num}.')



# Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import math

try:
    from collections.abc import Sequence
except Exception:
    from collections import Sequence

import logging
import cv2
import numpy as np
from antgo.dataflow.core import *
from .operators import BaseOperator
from .op_helper import jaccard_overlap, gaussian2D



logger = logging.getLogger(__name__)


class PadBatch(BaseOperator):
    """
    Pad a batch of samples so they can be divisible by a stride.
    The layout of each image should be 'CHW'.
    Args:
        pad_to_stride (int): If `pad_to_stride > 0`, pad zeros to ensure
            height and width is divisible by `pad_to_stride`.
    """

    def __init__(self, pad_to_stride=0, use_padded_im_info=True, inputs=None):
        super(PadBatch, self).__init__(inputs=inputs)
        self.pad_to_stride = pad_to_stride
        self.use_padded_im_info = use_padded_im_info

    def __call__(self, samples, context=None):
        """
        Args:
            samples (list): a batch of sample, each is dict.
        """
        coarsest_stride = self.pad_to_stride
        if coarsest_stride == 0:
            return samples
        max_shape = np.array([data['image'].shape for data in samples]).max(
            axis=0)

        if coarsest_stride > 0:
            max_shape[1] = int(
                np.ceil(max_shape[1] / coarsest_stride) * coarsest_stride)
            max_shape[2] = int(
                np.ceil(max_shape[2] / coarsest_stride) * coarsest_stride)

        padding_batch = []
        for data in samples:
            im = data['image']
            im_c, im_h, im_w = im.shape[:]
            padding_im = np.zeros(
                (im_c, max_shape[1], max_shape[2]), dtype=np.float32)
            padding_im[:, :im_h, :im_w] = im
            data['image'] = padding_im
            if self.use_padded_im_info:
                data['im_info'][:2] = max_shape[1:3]
            if 'semantic' in data.keys() and data['semantic'] is not None:
                semantic = data['semantic']
                padding_sem = np.zeros(
                    (1, max_shape[1], max_shape[2]), dtype=np.float32)
                padding_sem[:, :im_h, :im_w] = semantic
                data['semantic'] = padding_sem

        return samples


class RandomShape(BaseOperator):
    """
    Randomly reshape a batch. If random_inter is True, also randomly
    select one an interpolation algorithm [cv2.INTER_NEAREST, cv2.INTER_LINEAR,
    cv2.INTER_AREA, cv2.INTER_CUBIC, cv2.INTER_LANCZOS4]. If random_inter is
    False, use cv2.INTER_NEAREST.
    Args:
        sizes (list): list of int, random choose a size from these
        random_inter (bool): whether to randomly interpolation, defalut true.
    """

    def __init__(self, sizes=[], random_inter=False, resize_box=False, inputs=None):
        super(RandomShape, self).__init__(inputs=inputs)
        self.sizes = sizes
        self.random_inter = random_inter
        self.interps = [
            cv2.INTER_NEAREST,
            cv2.INTER_LINEAR,
            cv2.INTER_AREA,
            cv2.INTER_CUBIC,
            cv2.INTER_LANCZOS4,
        ] if random_inter else []
        self.resize_box = resize_box

    def __call__(self, samples, context=None):
        shape = np.random.choice(self.sizes)
        method = np.random.choice(self.interps) if self.random_inter \
            else cv2.INTER_NEAREST
        for i in range(len(samples)):
            im = samples[i]['image']
            h, w = im.shape[:2]
            scale_x = float(shape) / w
            scale_y = float(shape) / h
            im = cv2.resize(
                im, None, None, fx=scale_x, fy=scale_y, interpolation=method)
            samples[i]['image'] = im
            if self.resize_box and 'gt_bbox' in samples[i] and len(samples[0][
                    'gt_bbox']) > 0:
                scale_array = np.array([scale_x, scale_y] * 2, dtype=np.float32)
                samples[i]['gt_bbox'] = np.clip(samples[i]['gt_bbox'] *
                                                scale_array, 0,
                                                float(shape) - 1)
        return samples


class PadMultiScaleTest(BaseOperator):
    """
    Pad the image so they can be divisible by a stride for multi-scale testing.
 
    Args:
        pad_to_stride (int): If `pad_to_stride > 0`, pad zeros to ensure
            height and width is divisible by `pad_to_stride`.
    """

    def __init__(self, pad_to_stride=0, inputs=None):
        super(PadMultiScaleTest, self).__init__(inputs=inputs)
        self.pad_to_stride = pad_to_stride

    def __call__(self, samples, context=None):
        coarsest_stride = self.pad_to_stride
        if coarsest_stride == 0:
            return samples

        batch_input = True
        if not isinstance(samples, Sequence):
            batch_input = False
            samples = [samples]
        if len(samples) != 1:
            raise ValueError("Batch size must be 1 when using multiscale test, "
                             "but now batch size is {}".format(len(samples)))
        for i in range(len(samples)):
            sample = samples[i]
            for k in sample.keys():
                # hard code
                if k.startswith('image'):
                    im = sample[k]
                    im_c, im_h, im_w = im.shape
                    max_h = int(
                        np.ceil(im_h / coarsest_stride) * coarsest_stride)
                    max_w = int(
                        np.ceil(im_w / coarsest_stride) * coarsest_stride)
                    padding_im = np.zeros(
                        (im_c, max_h, max_w), dtype=np.float32)

                    padding_im[:, :im_h, :im_w] = im
                    sample[k] = padding_im
                    info_name = 'im_info' if k == 'image' else 'im_info_' + k
                    # update im_info
                    sample[info_name][:2] = [max_h, max_w]
        if not batch_input:
            samples = samples[0]
        return samples


class Gt2YoloTarget(BaseOperator):
    """
    Generate YOLOv3 targets by groud truth data, this operators is only used in
    fine grained YOLOv3 loss mode
    """

    def __init__(self,
                 anchors,
                 anchor_masks,
                 downsample_ratios,
                 num_classes=80,
                 iou_thresh=1.,
                 inputs=None):
        super(Gt2YoloTarget, self).__init__(inputs=inputs)
        self.anchors = anchors
        self.anchor_masks = anchor_masks
        self.downsample_ratios = downsample_ratios
        self.num_classes = num_classes
        self.iou_thresh = iou_thresh

    def __call__(self, samples, context=None):
        assert len(self.anchor_masks) == len(self.downsample_ratios), \
            "anchor_masks', and 'downsample_ratios' should have same length."

        h, w = samples[0]['image'].shape[1:3]
        an_hw = np.array(self.anchors) / np.array([[w, h]])
        for sample in samples:
            # im, gt_bbox, gt_class, gt_score = sample
            im = sample['image']
            gt_bbox = sample['gt_bbox']
            gt_class = sample['gt_class']
            gt_score = sample['gt_score']
            for i, (
                    mask, downsample_ratio
            ) in enumerate(zip(self.anchor_masks, self.downsample_ratios)):
                grid_h = int(h / downsample_ratio)
                grid_w = int(w / downsample_ratio)
                target = np.zeros(
                    (len(mask), 6 + self.num_classes, grid_h, grid_w),
                    dtype=np.float32)
                for b in range(gt_bbox.shape[0]):
                    gx, gy, gw, gh = gt_bbox[b, :]
                    cls = gt_class[b]
                    score = gt_score[b]
                    if gw <= 0. or gh <= 0. or score <= 0.:
                        continue

                    # find best match anchor index
                    best_iou = 0.
                    best_idx = -1
                    for an_idx in range(an_hw.shape[0]):
                        iou = jaccard_overlap(
                            [0., 0., gw, gh],
                            [0., 0., an_hw[an_idx, 0], an_hw[an_idx, 1]])
                        if iou > best_iou:
                            best_iou = iou
                            best_idx = an_idx

                    gi = int(gx * grid_w)
                    gj = int(gy * grid_h)

                    # gtbox should be regresed in this layes if best match 
                    # anchor index in anchor mask of this layer
                    if best_idx in mask:
                        best_n = mask.index(best_idx)

                        # x, y, w, h, scale
                        target[best_n, 0, gj, gi] = gx * grid_w - gi
                        target[best_n, 1, gj, gi] = gy * grid_h - gj
                        target[best_n, 2, gj, gi] = np.log(
                            gw * w / self.anchors[best_idx][0])
                        target[best_n, 3, gj, gi] = np.log(
                            gh * h / self.anchors[best_idx][1])
                        target[best_n, 4, gj, gi] = 2.0 - gw * gh

                        # objectness record gt_score
                        target[best_n, 5, gj, gi] = score

                        # classification
                        target[best_n, 6 + cls, gj, gi] = 1.

                    # For non-matched anchors, calculate the target if the iou 
                    # between anchor and gt is larger than iou_thresh
                    if self.iou_thresh < 1:
                        for idx, mask_i in enumerate(mask):
                            if mask_i == best_idx: continue
                            iou = jaccard_overlap(
                                [0., 0., gw, gh],
                                [0., 0., an_hw[mask_i, 0], an_hw[mask_i, 1]])
                            if iou > self.iou_thresh:
                                # x, y, w, h, scale
                                target[idx, 0, gj, gi] = gx * grid_w - gi
                                target[idx, 1, gj, gi] = gy * grid_h - gj
                                target[idx, 2, gj, gi] = np.log(
                                    gw * w / self.anchors[mask_i][0])
                                target[idx, 3, gj, gi] = np.log(
                                    gh * h / self.anchors[mask_i][1])
                                target[idx, 4, gj, gi] = 2.0 - gw * gh

                                # objectness record gt_score
                                target[idx, 5, gj, gi] = score

                                # classification
                                target[idx, 6 + cls, gj, gi] = 1.
                sample['target{}'.format(i)] = target
        return samples


class Gt2FCOSTarget(BaseOperator):
    """
    Generate FCOS targets by groud truth data
    """

    def __init__(self,
                 object_sizes_boundary,
                 center_sampling_radius,
                 downsample_ratios,
                 norm_reg_targets=False,
                 inputs=None):
        super(Gt2FCOSTarget, self).__init__(inputs=inputs)
        self.center_sampling_radius = center_sampling_radius
        self.downsample_ratios = downsample_ratios
        self.INF = np.inf
        self.object_sizes_boundary = [-1] + object_sizes_boundary + [self.INF]
        object_sizes_of_interest = []
        for i in range(len(self.object_sizes_boundary) - 1):
            object_sizes_of_interest.append([
                self.object_sizes_boundary[i], self.object_sizes_boundary[i + 1]
            ])
        self.object_sizes_of_interest = object_sizes_of_interest
        self.norm_reg_targets = norm_reg_targets

    def _compute_points(self, w, h):
        """
        compute the corresponding points in each feature map
        :param h: image height
        :param w: image width
        :return: points from all feature map
        """
        locations = []
        for stride in self.downsample_ratios:
            shift_x = np.arange(0, w, stride).astype(np.float32)
            shift_y = np.arange(0, h, stride).astype(np.float32)
            shift_x, shift_y = np.meshgrid(shift_x, shift_y)
            shift_x = shift_x.flatten()
            shift_y = shift_y.flatten()
            location = np.stack([shift_x, shift_y], axis=1) + stride // 2
            locations.append(location)
        num_points_each_level = [len(location) for location in locations]
        locations = np.concatenate(locations, axis=0)
        return locations, num_points_each_level

    def _convert_xywh2xyxy(self, gt_bbox, w, h):
        """
        convert the bounding box from style xywh to xyxy
        :param gt_bbox: bounding boxes normalized into [0, 1]
        :param w: image width
        :param h: image height
        :return: bounding boxes in xyxy style
        """
        bboxes = gt_bbox.copy()
        bboxes[:, [0, 2]] = bboxes[:, [0, 2]] * w
        bboxes[:, [1, 3]] = bboxes[:, [1, 3]] * h
        bboxes[:, 2] = bboxes[:, 0] + bboxes[:, 2]
        bboxes[:, 3] = bboxes[:, 1] + bboxes[:, 3]
        return bboxes

    def _check_inside_boxes_limited(self, gt_bbox, xs, ys,
                                    num_points_each_level):
        """
        check if points is within the clipped boxes
        :param gt_bbox: bounding boxes
        :param xs: horizontal coordinate of points
        :param ys: vertical coordinate of points
        :return: the mask of points is within gt_box or not
        """
        bboxes = np.reshape(
            gt_bbox, newshape=[1, gt_bbox.shape[0], gt_bbox.shape[1]])
        bboxes = np.tile(bboxes, reps=[xs.shape[0], 1, 1])
        ct_x = (bboxes[:, :, 0] + bboxes[:, :, 2]) / 2
        ct_y = (bboxes[:, :, 1] + bboxes[:, :, 3]) / 2
        beg = 0
        clipped_box = bboxes.copy()
        for lvl, stride in enumerate(self.downsample_ratios):
            end = beg + num_points_each_level[lvl]
            stride_exp = self.center_sampling_radius * stride
            clipped_box[beg:end, :, 0] = np.maximum(
                bboxes[beg:end, :, 0], ct_x[beg:end, :] - stride_exp)
            clipped_box[beg:end, :, 1] = np.maximum(
                bboxes[beg:end, :, 1], ct_y[beg:end, :] - stride_exp)
            clipped_box[beg:end, :, 2] = np.minimum(
                bboxes[beg:end, :, 2], ct_x[beg:end, :] + stride_exp)
            clipped_box[beg:end, :, 3] = np.minimum(
                bboxes[beg:end, :, 3], ct_y[beg:end, :] + stride_exp)
            beg = end
        l_res = xs - clipped_box[:, :, 0]
        r_res = clipped_box[:, :, 2] - xs
        t_res = ys - clipped_box[:, :, 1]
        b_res = clipped_box[:, :, 3] - ys
        clipped_box_reg_targets = np.stack([l_res, t_res, r_res, b_res], axis=2)
        inside_gt_box = np.min(clipped_box_reg_targets, axis=2) > 0
        return inside_gt_box

    def __call__(self, samples, context=None):
        assert len(self.object_sizes_of_interest) == len(self.downsample_ratios), \
            "object_sizes_of_interest', and 'downsample_ratios' should have same length."

        for sample in samples:
            # im, gt_bbox, gt_class, gt_score = sample
            im = sample['image']
            im_info = sample['im_info']
            bboxes = sample['gt_bbox']
            gt_class = sample['gt_class']
            gt_score = sample['gt_score']
            bboxes[:, [0, 2]] = bboxes[:, [0, 2]] * np.floor(im_info[1]) / \
                np.floor(im_info[1] / im_info[2])
            bboxes[:, [1, 3]] = bboxes[:, [1, 3]] * np.floor(im_info[0]) / \
                np.floor(im_info[0] / im_info[2])
            # calculate the locations
            h, w = sample['image'].shape[1:3]
            points, num_points_each_level = self._compute_points(w, h)
            object_scale_exp = []
            for i, num_pts in enumerate(num_points_each_level):
                object_scale_exp.append(
                    np.tile(
                        np.array([self.object_sizes_of_interest[i]]),
                        reps=[num_pts, 1]))
            object_scale_exp = np.concatenate(object_scale_exp, axis=0)

            gt_area = (bboxes[:, 2] - bboxes[:, 0]) * (
                bboxes[:, 3] - bboxes[:, 1])
            xs, ys = points[:, 0], points[:, 1]
            xs = np.reshape(xs, newshape=[xs.shape[0], 1])
            xs = np.tile(xs, reps=[1, bboxes.shape[0]])
            ys = np.reshape(ys, newshape=[ys.shape[0], 1])
            ys = np.tile(ys, reps=[1, bboxes.shape[0]])

            l_res = xs - bboxes[:, 0]
            r_res = bboxes[:, 2] - xs
            t_res = ys - bboxes[:, 1]
            b_res = bboxes[:, 3] - ys
            reg_targets = np.stack([l_res, t_res, r_res, b_res], axis=2)
            if self.center_sampling_radius > 0:
                is_inside_box = self._check_inside_boxes_limited(
                    bboxes, xs, ys, num_points_each_level)
            else:
                is_inside_box = np.min(reg_targets, axis=2) > 0
            # check if the targets is inside the corresponding level
            max_reg_targets = np.max(reg_targets, axis=2)
            lower_bound = np.tile(
                np.expand_dims(
                    object_scale_exp[:, 0], axis=1),
                reps=[1, max_reg_targets.shape[1]])
            high_bound = np.tile(
                np.expand_dims(
                    object_scale_exp[:, 1], axis=1),
                reps=[1, max_reg_targets.shape[1]])
            is_match_current_level = \
                (max_reg_targets > lower_bound) & \
                (max_reg_targets < high_bound)
            points2gtarea = np.tile(
                np.expand_dims(
                    gt_area, axis=0), reps=[xs.shape[0], 1])
            points2gtarea[is_inside_box == 0] = self.INF
            points2gtarea[is_match_current_level == 0] = self.INF
            points2min_area = points2gtarea.min(axis=1)
            points2min_area_ind = points2gtarea.argmin(axis=1)
            labels = gt_class[points2min_area_ind] + 1
            labels[points2min_area == self.INF] = 0
            reg_targets = reg_targets[range(xs.shape[0]), points2min_area_ind]
            ctn_targets = np.sqrt((reg_targets[:, [0, 2]].min(axis=1) / \
                                  reg_targets[:, [0, 2]].max(axis=1)) * \
                                  (reg_targets[:, [1, 3]].min(axis=1) / \
                                   reg_targets[:, [1, 3]].max(axis=1))).astype(np.float32)
            ctn_targets = np.reshape(
                ctn_targets, newshape=[ctn_targets.shape[0], 1])
            ctn_targets[labels <= 0] = 0
            pos_ind = np.nonzero(labels != 0)
            reg_targets_pos = reg_targets[pos_ind[0], :]
            split_sections = []
            beg = 0
            for lvl in range(len(num_points_each_level)):
                end = beg + num_points_each_level[lvl]
                split_sections.append(end)
                beg = end
            labels_by_level = np.split(labels, split_sections, axis=0)
            reg_targets_by_level = np.split(reg_targets, split_sections, axis=0)
            ctn_targets_by_level = np.split(ctn_targets, split_sections, axis=0)
            for lvl in range(len(self.downsample_ratios)):
                grid_w = int(np.ceil(w / self.downsample_ratios[lvl]))
                grid_h = int(np.ceil(h / self.downsample_ratios[lvl]))
                if self.norm_reg_targets:
                    sample['reg_target{}'.format(lvl)] = \
                        np.reshape(
                            reg_targets_by_level[lvl] / \
                            self.downsample_ratios[lvl],
                            newshape=[grid_h, grid_w, 4])
                else:
                    sample['reg_target{}'.format(lvl)] = np.reshape(
                        reg_targets_by_level[lvl],
                        newshape=[grid_h, grid_w, 4])
                sample['labels{}'.format(lvl)] = np.reshape(
                    labels_by_level[lvl], newshape=[grid_h, grid_w, 1])
                sample['centerness{}'.format(lvl)] = np.reshape(
                    ctn_targets_by_level[lvl], newshape=[grid_h, grid_w, 1])
        return samples


class Gt2TTFTarget(BaseOperator):
    """
    Gt2TTFTarget
    Generate TTFNet targets by ground truth data
    
    Args:
        num_classes(int): the number of classes.
        down_ratio(int): the down ratio from images to heatmap, 4 by default.
        alpha(float): the alpha parameter to generate gaussian target.
            0.54 by default.
    """

    def __init__(self, num_classes, down_ratio=4, alpha=0.54, inputs=None):
        super(Gt2TTFTarget, self).__init__(inputs=inputs)
        self.down_ratio = down_ratio
        self.num_classes = num_classes
        self.alpha = alpha

    def __call__(self, samples, context=None):
        output_size = samples[0]['image'].shape[1]
        feat_size = output_size // self.down_ratio
        for sample in samples:
            heatmap = np.zeros(
                (self.num_classes, feat_size, feat_size), dtype='float32')
            box_target = np.ones(
                (4, feat_size, feat_size), dtype='float32') * -1
            reg_weight = np.zeros((1, feat_size, feat_size), dtype='float32')

            '''
            if "gt_keypoint" in sample:
                continue
            '''
            gt_bbox = sample['gt_bbox']
            gt_class = sample['gt_class']

            bbox_w = gt_bbox[:, 2] - gt_bbox[:, 0] + 1
            bbox_h = gt_bbox[:, 3] - gt_bbox[:, 1] + 1
            area = bbox_w * bbox_h
            boxes_areas_log = np.log(area)
            boxes_ind = np.argsort(boxes_areas_log, axis=0)[::-1]
            boxes_area_topk_log = boxes_areas_log[boxes_ind]
            gt_bbox = gt_bbox[boxes_ind]
            gt_class = gt_class[boxes_ind]

            feat_gt_bbox = gt_bbox / self.down_ratio
            feat_gt_bbox = np.clip(feat_gt_bbox, 0, feat_size - 1)
            feat_hs, feat_ws = (feat_gt_bbox[:, 3] - feat_gt_bbox[:, 1],
                                feat_gt_bbox[:, 2] - feat_gt_bbox[:, 0])

            ct_inds = np.stack(
                [(gt_bbox[:, 0] + gt_bbox[:, 2]) / 2,
                 (gt_bbox[:, 1] + gt_bbox[:, 3]) / 2],
                axis=1) / self.down_ratio

            h_radiuses_alpha = (feat_hs / 2. * self.alpha).astype('int32')
            w_radiuses_alpha = (feat_ws / 2. * self.alpha).astype('int32')

            for k in range(len(gt_bbox)):
                cls_id = gt_class[k]
                fake_heatmap = np.zeros((feat_size, feat_size), dtype='float32')
                self.draw_truncate_gaussian(fake_heatmap, ct_inds[k],
                                            h_radiuses_alpha[k],
                                            w_radiuses_alpha[k])

                heatmap[cls_id] = np.maximum(heatmap[cls_id], fake_heatmap)
                box_target_inds = fake_heatmap > 0
                box_target[:, box_target_inds] = gt_bbox[k][:, None]

                local_heatmap = fake_heatmap[box_target_inds]
                ct_div = np.sum(local_heatmap)
                local_heatmap *= boxes_area_topk_log[k]
                reg_weight[0, box_target_inds] = local_heatmap / ct_div
            sample['ttf_heatmap'] = heatmap
            sample['ttf_box_target'] = box_target
            sample['ttf_reg_weight'] = reg_weight
        return samples

    def draw_truncate_gaussian(self, heatmap, center, h_radius, w_radius):
        h, w = 2 * h_radius + 1, 2 * w_radius + 1
        sigma_x = w / 6
        sigma_y = h / 6
        gaussian = gaussian2D((h, w), sigma_x, sigma_y)

        x, y = int(center[0]), int(center[1])

        height, width = heatmap.shape[0:2]

        left, right = min(x, w_radius), min(width - x, w_radius + 1)
        top, bottom = min(y, h_radius), min(height - y, h_radius + 1)

        masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
        masked_gaussian = gaussian[h_radius - top:h_radius + bottom, w_radius -
                                   left:w_radius + right]
        if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:
            heatmap[y - top:y + bottom, x - left:x + right] = np.maximum(
                masked_heatmap, masked_gaussian)
        return heatmap


class OffsetGenerator():
    def __init__(self, output_h, output_w, num_joints, radius):
        self.num_joints_without_center = num_joints
        self.num_joints_with_center = num_joints+1
        self.output_w = output_w
        self.output_h = output_h
        self.radius = radius

    def __call__(self, joints, area):
        assert joints.shape[1] == self.num_joints_with_center, \
            'the number of joints should be 18, 17 keypoints + 1 center joint.'

        offset_map = np.zeros((self.num_joints_without_center*2, self.output_h, self.output_w),
                              dtype=np.float32)
        weight_map = np.zeros((self.num_joints_without_center*2, self.output_h, self.output_w),
                              dtype=np.float32)

        area_map = np.zeros((self.output_h, self.output_w),
                            dtype=np.float32)

        for person_id, p in enumerate(joints):
            ct_x = int(p[-1, 0])
            ct_y = int(p[-1, 1])
            ct_v = int(p[-1, 2])
            if ct_v < 1 or ct_x < 0 or ct_y < 0 \
                    or ct_x >= self.output_w or ct_y >= self.output_h:
                continue

            for idx, pt in enumerate(p[:-1]):
                if pt[2] > 0:
                    x, y = pt[0], pt[1]
                    if x < 0 or y < 0 or \
                            x >= self.output_w or y >= self.output_h:
                        continue

                    start_x = max(int(ct_x - self.radius), 0)
                    start_y = max(int(ct_y - self.radius), 0)
                    end_x = min(int(ct_x + self.radius), self.output_w)
                    end_y = min(int(ct_y + self.radius), self.output_h)

                    for pos_x in range(start_x, end_x):
                        for pos_y in range(start_y, end_y):
                            offset_x = pos_x - x
                            offset_y = pos_y - y
                            if offset_map[idx*2, pos_y, pos_x] != 0 \
                                    or offset_map[idx*2+1, pos_y, pos_x] != 0:
                                if area_map[pos_y, pos_x] < area[person_id]:
                                    continue
                                
                            offset_map[idx*2, pos_y, pos_x] = offset_x          # normalize
                            offset_map[idx*2+1, pos_y, pos_x] = offset_y
                            weight_map[idx*2, pos_y, pos_x] = 1. / (np.sqrt(area[person_id]) + 1e-6)
                            weight_map[idx*2+1, pos_y, pos_x] = 1. / (np.sqrt(area[person_id]) + 1e-6)
                            area_map[pos_y, pos_x] = area[person_id]

        return offset_map, weight_map

        

class PafWithCenterGenerator():
    def __init__(self, output_h, output_w, num_joints, width=1.0, is_reduce=True):
        self.output_h = output_h
        self.output_w = output_w
        self.width = width
        self.num_joints = num_joints

    def __call__(self, joints, area):
        # ignore area
        feat_size_h = self.output_h
        feat_size_w = self.output_w
        heatmap_paf = np.zeros((feat_size_h, feat_size_w, 2), dtype='float32')
        count = np.zeros((int(feat_size_h), int(feat_size_w)), dtype=np.uint32)

        for person_id, p in enumerate(joints):
            ct_x, ct_y, ct_v = p[ -1, :3]
            if (ct_v < 1 or ct_x < 0 or ct_y < 0 \
                    or ct_x >= self.output_w or ct_y >= self.output_h):
                continue

            for idx, pt in enumerate(p[:-1]):
                heatmap_paf, count = putVecMaps(
                    centerA=p[-1,:2],
                    centerB=pt[:2],
                    accumulate_vec_map=heatmap_paf,
                    count=count,
                    grid_y=feat_size_h, grid_x=feat_size_w,
                    stride=1,
                    width=self.width
                )

        heatmap_paf = np.transpose(heatmap_paf, (2,0,1))
        return heatmap_paf


class Gt2TTFKPTTarget(Gt2TTFTarget):
    """
    Gt2TTFTarget
    Generate TTFNet targets by ground truth data
    
    Args:
        num_classes(int): the number of classes.
        down_ratio(int): the down ratio from images to heatmap, 4 by default.
        alpha(float): the alpha parameter to generate gaussian target.
            0.54 by default.
    """

    def __init__(self, 
                num_classes, 
                down_ratio=4, 
                alpha=0.54, 
                num_kpts=21, 
                has_offset=False,
                has_paf=False, 
                kpt_down_scale=2, 
                inputs=None):
        super(Gt2TTFKPTTarget, self).__init__(
                num_classes=num_classes,
                down_ratio=down_ratio,
                alpha=alpha,
                inputs=inputs)
        self.num_kpts = num_kpts
        self.offset_generator = None
        if has_offset:
            self.offset_generator = OffsetGenerator(160//4, 192//4, 21, 4)
        
        self.paf_generator = None
        if has_paf:
            self.paf_generator = \
                PafWithCenterGenerator(output_h=160//4, 
                                       output_w=192//4,
                                       num_joints=21, 
                                       width=1.0)
        
        self.kpt_down_scale = kpt_down_scale
        
    def __call__(self, samples, context=None):
        # output_size = samples[0]['image'].shape[1]
        output_size_h, output_size_w = samples[0]['image'].shape[1:3]
        feat_size_h = output_size_h // self.down_ratio
        feat_size_w = output_size_w // self.down_ratio
        heatmap_kpt_h = output_size_h // self.kpt_down_scale
        heatmap_kpt_w = output_size_w // self.kpt_down_scale
        
        count = 0
        for sample in samples:
            heatmap = np.zeros(
                (self.num_classes, feat_size_h, feat_size_w), dtype='float32')
            heatmap_kpt = np.zeros(
                (self.num_kpts, heatmap_kpt_h, heatmap_kpt_w), dtype='float32')            
            heatmap_one = np.zeros(
                (1, feat_size_h, feat_size_w), dtype='float32')
            box_target = np.ones(
                (4, feat_size_h, feat_size_w), dtype='float32') * -1
            reg_weight = np.zeros((1, feat_size_h, feat_size_w), dtype='float32')     

            # heatmap mask valid
            heatmap_mask = np.ones((1, feat_size_h, feat_size_w), dtype='float32')

            # henggang kpt
            henggang_kpt = np.zeros((4, heatmap_kpt_h, heatmap_kpt_w), dtype='float32')

            gt_bbox = sample['gt_bbox']
            gt_class = sample['gt_class']
            bbox_w = gt_bbox[:, 2] - gt_bbox[:, 0] + 1
            bbox_h = gt_bbox[:, 3] - gt_bbox[:, 1] + 1
            area = bbox_w * bbox_h
            boxes_areas_log = np.log(area)
            # boxes_ind = np.argsort(boxes_areas_log, axis=0)[::-1]
            # boxes_area_topk_log = boxes_areas_log[boxes_ind]
            # gt_bbox = gt_bbox[boxes_ind]
            # gt_class = gt_class[boxes_ind]

            feat_gt_bbox = gt_bbox / self.down_ratio
            # feat_gt_bbox = np.clip(feat_gt_bbox, 0, feat_size - 1)

            feat_hs, feat_ws = (feat_gt_bbox[:, 3] - feat_gt_bbox[:, 1],
                                feat_gt_bbox[:, 2] - feat_gt_bbox[:, 0])

            ct_inds = np.stack(
                [(gt_bbox[:, 0] + gt_bbox[:, 2]) / 2,
                 (gt_bbox[:, 1] + gt_bbox[:, 3]) / 2],
                axis=1) / self.down_ratio
            
            h_radiuses_alpha = (feat_hs / 2. * self.alpha + 0.5).astype('int32')
            w_radiuses_alpha = (feat_ws / 2. * self.alpha + 0.5).astype('int32')
            # boxes_scale = (feat_gt_bbox[:, 3] - feat_gt_bbox[:, 1]) * \
            #                   (feat_gt_bbox[:, 2] - feat_gt_bbox[:, 0]) / \
            #                   ((feat_size_w - 1) * (feat_size_h - 1))

            # num_obj = ct_inds.shape[0]
            # ct_inds_ext = np.concatenate([ct_inds, np.ones((num_obj, 1))], axis=1)
            # joints = np.zeros((num_ins, self.num_kpts, 3))
            # joints = np.concatenate([joints, np.expand_dims(ct_inds_ext, 1)], axis=1)       # num_ins, self.num_kpts+1,3
            # person_areas = (feat_gt_bbox[:, 3] - feat_gt_bbox[:, 1]) * (feat_gt_bbox[:, 2] - feat_gt_bbox[:, 0])
            
            body_bbox = []
            person_num = 0
            for k in range(len(gt_bbox)):
                cls_id = gt_class[k]
                if cls_id == 0:             # 人体类别 0
                    person_num += 1
                    body_bbox.append(gt_bbox[k].copy())
                
                fake_heatmap = np.zeros((feat_size_h, feat_size_w), dtype='float32')
                self.draw_truncate_gaussian(fake_heatmap, ct_inds[k], h_radiuses_alpha[k], w_radiuses_alpha[k])

                heatmap[cls_id] = np.maximum(heatmap[cls_id], fake_heatmap)
                box_target_inds = fake_heatmap > 0
                box_target[:, box_target_inds] = gt_bbox[k][:, None]

                local_heatmap = fake_heatmap[box_target_inds]
                ct_div = np.sum(local_heatmap)
                local_heatmap *= boxes_areas_log[k]
                reg_weight[0, box_target_inds] = local_heatmap / ct_div

            joints = np.zeros((person_num, self.num_kpts+1, 3))
            person_areas = np.zeros((person_num))
            if 'gt_keypoint' in sample:
                gt_kpts = sample['gt_keypoint'].copy()
                gt_kpts_for_heatmap = sample['gt_keypoint'].copy()
                gt_kpts[:,:,:2] = gt_kpts[:,:,:2] / self.down_ratio
                gt_kpts_for_heatmap[:,:,:2] = gt_kpts_for_heatmap[:,:,:2] / self.kpt_down_scale
                gt_kpts_from_bbox_i = sample['gt_keypoint_from_bbox_i']
                
                num_ins = gt_kpts.shape[0]                
                joints[:, :-1, :] = gt_kpts
                
                # 计算每个人的中心
                for k in range(num_ins):
                    from_bbox_i = gt_kpts_from_bbox_i[k]
                    joints[k, -1, 0] = (feat_gt_bbox[from_bbox_i, 0] + feat_gt_bbox[from_bbox_i, 2])/2.0
                    joints[k, -1, 1] = (feat_gt_bbox[from_bbox_i, 1] + feat_gt_bbox[from_bbox_i, 3])/2.0
                    joints[k, -1, 2] = 1.0                    
                
                # 计算每个人的面积
                for k in range(num_ins):
                    person_min_x = 10000
                    person_max_x = 0
                    person_min_y = 10000
                    person_max_y = 0
                    for jter in range(self.num_kpts):
                        if gt_kpts[k, jter, 2] < 0.001:
                            # 忽略无效关键点
                            continue
                        
                        person_min_x = min(gt_kpts[k, jter, 0], person_min_x)
                        person_max_x = max(gt_kpts[k, jter, 0], person_max_x)
                        
                        person_min_y = min(gt_kpts[k, jter, 1], person_min_y)
                        person_max_y = max(gt_kpts[k, jter, 1], person_max_y)
                    
                    person_areas[k] = max(0, (person_max_x-person_min_x)*(person_max_y-person_min_y))
                        
                # # 构建KPT heatmap
                # # DEBUG:
                # ssss = sample['image'].copy()
                # ssss = np.transpose(ssss, (1,2,0)).astype(np.uint8).copy()
                for k in range(num_ins):
                    for jter in range(self.num_kpts):
                        if gt_kpts_for_heatmap[k, jter, 2] < 0.001:
                            # 忽略无效关键点
                            continue
                        
                        fake_heatmap = np.zeros((heatmap_kpt_h, heatmap_kpt_w), dtype='float32')
                        kpt_pt = gt_kpts_for_heatmap[k, jter, 0:2]
                        # 这里kpt采用固定宽度构建heatmap
                        self.draw_truncate_gaussian(fake_heatmap, kpt_pt, 7, 7)
                        heatmap_kpt[jter] = np.maximum(heatmap_kpt[jter], fake_heatmap)
                        
                        # # DEBUG: 
                        # xx = min((int)(gt_kpts_for_heatmap[k, jter, 0]) * self.kpt_down_scale, 191)
                        # yy = min((int)(gt_kpts_for_heatmap[k, jter, 1]) * self.kpt_down_scale, 159)
                        # cv2.circle(ssss, (xx,yy), 5, (0,0,255))
                        # cv2.putText(ssss, 
                        #                 f"{jter}", 
                        #                 (xx-1, yy-5), 
                        #                 cv2.FONT_HERSHEY_SIMPLEX, 
                        #                 0.3,
                        #                 (0, 255, 0),
                        #                 1)

                heatmap_kpt = np.clip(heatmap_kpt, a_min=0.0, a_max=1.0)
                
                # # # DEBUG: 显示heatmap kpt
                # # cv2.imwrite(f"./temp/s_id_{count}.png", (np.max(heatmap_kpt, 0)*255).astype(np.uint8))
                # cv2.imwrite(f"./temp/s_image_id_{count}.png", ssss)

            sample['ttf_heatmap'] = heatmap
            sample['ttf_kpt_heatmap'] = heatmap_kpt
            sample['ttf_box_target'] = box_target
            sample['ttf_reg_weight'] = reg_weight
            sample['ttf_ext_kpt_heatmap'] = henggang_kpt
            sample['ttf_heatmap_mask'] = heatmap_mask

            # cv2.imwrite(f"./temp/{count}_left.png", (henggang_kpt[0]*255).astype(np.uint8))
            # cv2.imwrite(f"./temp/{count}_right.png", (henggang_kpt[1]*255).astype(np.uint8))
            body_bbox = np.stack(body_bbox)
            body_bbox = body_bbox.astype(np.float32)
            half_body_bbox_w = (body_bbox[:, 2] - body_bbox[:, 0] + 1)/2.0
            half_body_bbox_h = (body_bbox[:, 3] - body_bbox[:, 1] + 1)/2.0            

            # 引入随机抖动half_size
            half_body_bbox_h = half_body_bbox_h * (np.random.random() * 0.5 + 0.8)
            half_body_bbox_w =  half_body_bbox_w * (np.random.random() * 0.5 + 0.8)
            body_bbox_x = (body_bbox[:, 0] + body_bbox[:, 2]) / 2
            body_bbox_y = (body_bbox[:, 1] + body_bbox[:, 3]) / 2
            # 引入随机抖动cx
            body_bbox_x += (2*np.random.random() - 1.0) * 0.3 * np.minimum(half_body_bbox_w,half_body_bbox_h)
            body_bbox_y += (2*np.random.random() - 1.0) * 0.3 * np.minimum(half_body_bbox_w,half_body_bbox_h)

            body_bbox = [body_bbox_x-half_body_bbox_w, 
                            body_bbox_y-half_body_bbox_h,
                            body_bbox_x+half_body_bbox_w,
                            body_bbox_y+half_body_bbox_h]
            body_bbox = np.stack(body_bbox, 1)

            # 确保固定包含3个框
            body_bbox_num = body_bbox.shape[0]
            body_bbox_select_list = list(range(body_bbox_num))
            np.random.shuffle(body_bbox_select_list)
            if len(body_bbox_select_list) < 3:
                add_list = list(np.random.choice(body_bbox_select_list, 3-len(body_bbox_select_list)))
                body_bbox_select_list = body_bbox_select_list + add_list
            body_bbox_select_list = body_bbox_select_list[:3]

            sample['ttf_roi_box'] = body_bbox[body_bbox_select_list].astype(np.float32)
            sample['ttf_roi_num'] = np.array([len(body_bbox_select_list)])

            sample['ttf_roi_kpt'] = np.zeros((len(body_bbox_select_list), 21, 40, 40), dtype=np.float32)    # Human x 21 x H x W
            gt_kpts_for_heatmap = sample['gt_keypoint'].copy()
            gt_kpts_for_heatmap[:,:,:2] = gt_kpts_for_heatmap[:,:,:2] / self.kpt_down_scale
            gt_kpts_from_bbox_i = sample['gt_keypoint_from_bbox_i']
            
            ttf_roi_kpt_offset = np.zeros((len(body_bbox_select_list), 21, 40, 40), dtype=np.float32)    # Human x 21 x H x W
            # 
            gt_kpts_reg_weight = np.zeros((len(body_bbox_select_list),1, 40, 40), dtype='float32') 
            gt_kpts_reg = np.ones((len(body_bbox_select_list), 2, 40, 40), dtype='float32') * -1

            for index, human_id in enumerate(body_bbox_select_list):
                human_x0,human_y0,human_x1,human_y1 = body_bbox[human_id] / self.kpt_down_scale
                human_w = human_x1 - human_x0
                human_h = human_y1 - human_y0
                
                for jter in range(self.num_kpts):
                    if gt_kpts_for_heatmap[human_id, jter, 2] < 0.001 or \
                        (gt_kpts_for_heatmap[human_id, jter, 0] < human_x0 or  gt_kpts_for_heatmap[human_id, jter, 1] < human_y0) or \
                        (gt_kpts_for_heatmap[human_id, jter, 0] > human_x1 or  gt_kpts_for_heatmap[human_id, jter, 1] > human_y1):
                        # 忽略无效关键点
                        continue
                    
                    kpt_pt = gt_kpts_for_heatmap[human_id, jter, 0:2]
                    joint_kpt_x, joint_kpt_y = kpt_pt
                    joint_kpt_x = (joint_kpt_x - human_x0) / human_w * 40.0
                    joint_kpt_y = (joint_kpt_y - human_y0) / human_h * 40.0
                    kpt_pt = np.array([joint_kpt_x, joint_kpt_y])

                    # for heatmap
                    fake_heatmap = np.zeros((40, 40), dtype='float32')                    
                    self.draw_truncate_gaussian(fake_heatmap, kpt_pt, 3, 3)
                    sample['ttf_roi_kpt'][index, jter] = np.maximum(sample['ttf_roi_kpt'][index, jter], fake_heatmap)

                    # for offset
                    fake_heatmap = np.zeros((40, 40), dtype='float32')
                    self.draw_truncate_gaussian(fake_heatmap, kpt_pt, 5, 5)
                    ttf_roi_kpt_offset[index, jter] = np.maximum(ttf_roi_kpt_offset[index, jter], fake_heatmap)

                    # float coordinate
                    inds = fake_heatmap < ttf_roi_kpt_offset[index, jter]
                    fake_heatmap[inds] = 0
                    inds = fake_heatmap > 0
                    mm = gt_kpts_reg[index]
                    mm[:, inds] = kpt_pt[:, None]
                    gt_kpts_reg[index] = mm
                
                gt_kpts_reg_weight[index, 0] = np.max(ttf_roi_kpt_offset[index], 0)

                # image = sample['image'].copy()
                # image = np.transpose(image, (1,2,0)).astype(np.uint8).copy()
                # human_x0 = np.clip((int)(human_x0) * 2, 0, image.shape[1])
                # human_y0 = np.clip((int)(human_y0) * 2, 0, image.shape[0])
                # human_x1 = np.clip((int)(human_x1) * 2, 0, image.shape[1])
                # human_y1 = np.clip((int)(human_y1) * 2, 0, image.shape[0])
                # human_image = image[(human_y0):(human_y1), (human_x0):(human_x1)]
                # cv2.imwrite('./human_image.jpg', human_image)
                # cv2.imwrite(f"./human_kpt.jpg", (np.max(sample['ttf_roi_kpt'][index], 0)*255).astype(np.uint8))

                # x0,y0,x1,y1 = sample['ttf_roi_box'][index]
                # human_x0 = np.clip((int)(x0), 0, image.shape[1])
                # human_y0 = np.clip((int)(y0), 0, image.shape[0])
                # human_x1 = np.clip((int)(x1), 0, image.shape[1])
                # human_y1 = np.clip((int)(y1), 0, image.shape[0])
                # human_image = image[(human_y0):(human_y1), (human_x0):(human_x1)]
                # cv2.imwrite('./human_image2.jpg', human_image)

            sample['ttf_roi_kpt'] = np.clip(sample['ttf_roi_kpt'], a_min=0.0, a_max=1.0)
            sample['ttf_roi_kpt_reg'] = gt_kpts_reg
            sample['ttf_roi_kpt_reg_weight'] = gt_kpts_reg_weight

            if self.offset_generator is not None:
                offset_map, weight_map = self.offset_generator(joints, person_areas)
                sample['ttf_offset_map'] = offset_map
                sample['ttf_offset_weight'] = weight_map            
            if self.paf_generator is not None:
                paf_map = self.paf_generator(joints, person_areas)
                sample['ttf_paf_map']= paf_map
                sample['ttf_paf_weight'] = np.ones(paf_map.shape)
            
            count += 1
            
        return samples

    def draw_truncate_gaussian(self, heatmap, center, h_radius, w_radius, factor=6., scale=0.0, omega=0.3):
        h, w = 2 * h_radius + 1, 2 * w_radius + 1
        sigma_x = w / factor * (1 - scale ** omega)
        sigma_y = h / factor * (1 - scale ** omega)
        gaussian = gaussian2D((h, w), sigma_x, sigma_y)

        x, y = int(center[0]+0.5), int(center[1]+0.5)

        height, width = heatmap.shape[0:2]

        left, right = min(x, w_radius), min(width - x, w_radius + 1)
        top, bottom = min(y, h_radius), min(height - y, h_radius + 1)

        masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
        masked_gaussian = gaussian[h_radius - top:h_radius + bottom, w_radius -
                                   left:w_radius + right]
        if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:
            heatmap[y - top:y + bottom, x - left:x + right] = np.maximum(
                masked_heatmap, masked_gaussian)
        return heatmap



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:01
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from .data_collection import DataCollection, DataFrame
from .entity import Entity
from .image import *
from .common import *
from antgo.pipeline.hparam import HyperParameter as State

from antgo.pipeline.hparam import param_scope
from antgo.pipeline.hparam import dynamic_dispatch
from antgo.pipeline.functional.common.config import *
import numpy as np

read_camera = DataCollection.read_camera
read_video = DataCollection.read_video


@dynamic_dispatch
def glob(*arg):  # pragma: no cover
    """
    Return a DataCollection of paths matching a pathname pattern.
    """

    index = param_scope()._index
    if index is None:
        return DataCollection.from_glob(*arg)
    return DataFrame.from_glob(*arg).map(lambda x: Entity(**{index: x}))


@dynamic_dispatch
def video_dc(*args):
    index = param_scope()._index
    if index is None:
        return DataCollection.read_video(*args)
      
    return DataFrame.read_video(*args).map(lambda x: Entity(**{index: x}))


@dynamic_dispatch
def camera_dc(*args):
    index = param_scope()._index
    if index is None:
        return DataCollection.read_camera(*args)
      
    return DataFrame.read_camera(*args).map(lambda x: Entity(**{index: x}))


def _api():
  """
  Create an API input, for building RestFul API or application API.
  """
  return DataFrame.api(index=param_scope()._index)


api = dynamic_dispatch(_api)


def _web():
  return DataFrame.web(index=param_scope()._index)


web = dynamic_dispatch(_web)


def _dc(iterable):
  """
  Return a DataCollection.
  """

  index = param_scope()._index
  if index is None:
    return DataCollection(iterable)
  if isinstance(index, (list, tuple)):
    return DataFrame(iterable).map(lambda x: Entity(**dict(zip(index, x))))
  return DataFrame(iterable).map(lambda x: Entity(**{index: x}))


dc = dynamic_dispatch(_dc)


def _placeholder(*arg):
  index = param_scope()._index

  # 
  if isinstance(index, tuple):
    for ii,xx in zip(index, arg):
      
      data_type = -1
      if xx.dtype == np.float32:
        data_type = 6
      elif xx.dtype == np.float64:
        data_type = 7
      elif xx.dtype == np.int32:
        data_type = 4
      elif xx.dtype == np.uint8:
        data_type = 1

      if data_type < 0:
        print('placeholder type abnormal.')
      
      add_op_info(
        'placeholder_op', 
        (None, (ii,)), 
        (), 
        {
          'memory_type': 2,     # CPU_BUFFER
          'data_format': 1000,  # AUTO
          'data_type': data_type,        # EAGLEEYE_UCHAR, EAGLEEYE_FLOAT
          'shape': list(xx.shape),  # 
        }
      )

    return DataFrame.placeholder(*arg).map(lambda x: Entity(**{ii: xx for ii, xx in zip(index, x) }))
  else:
    for ii, xx in zip([index], arg):
      data_type = -1
      if xx.dtype == np.float32:
        data_type = 6
      elif xx.dtype == np.int32:
        data_type = 4
      elif xx.dtype == np.uint8:
        data_type = 1

      if data_type < 0:
        print('placeholder type abnormal.')
      
      add_op_info(
        'placeholder_op', 
        (None, (ii,)), 
        (), 
        {
          'memory_type': 2,               # CPU_BUFFER
          'data_format': 1000,            # AUTO
          'data_type': data_type,         # EAGLEEYE_UCHAR, EAGLEEYE_FLOAT
          'shape': list(xx.shape),        # 
        }
      )

    return DataFrame.placeholder(*arg).map(lambda x: Entity(**{index: x }))

placeholder = dynamic_dispatch(_placeholder)



# -*- coding: UTF-8 -*-
# @Time    : 2022/5/2 19:49
# @File    : utils.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import torch

PADDLE_TORCH_TYPE_MAPPER = {
  "float32": torch.float32,
  "float64": torch.float64,
  "int64": torch.int64,
  "int32": torch.int32,
}


class CPUPlace(object):
  def __init__(self):
    pass

  def device(self):
    return torch.device('cpu')

class CUDAPlace(object):
  def __init__(self, cuda_id=0):
    self.cuda_id = cuda_id

  def device(self):
    return torch.device(f'cuda:{self.cuda_id}')


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class FSP(nn.Module):
	'''
	A Gift from Knowledge Distillation: Fast Optimization, Network Minimization and Transfer Learning
	http://openaccess.thecvf.com/content_cvpr_2017/papers/Yim_A_Gift_From_CVPR_2017_paper.pdf
	'''
	def __init__(self):
		super(FSP, self).__init__()

	def forward(self, fm_s1, fm_s2, fm_t1, fm_t2):
		loss = F.mse_loss(self.fsp_matrix(fm_s1,fm_s2), self.fsp_matrix(fm_t1,fm_t2))

		return loss

	def fsp_matrix(self, fm1, fm2):
		if fm1.size(2) > fm2.size(2):
			fm1 = F.adaptive_avg_pool2d(fm1, (fm2.size(2), fm2.size(3)))

		fm1 = fm1.view(fm1.size(0), fm1.size(1), -1)
		fm2 = fm2.view(fm2.size(0), fm2.size(1), -1).transpose(1,2)

		fsp = torch.bmm(fm1, fm2) / fm1.size(2)

		return fsp


"""Building Utils"""

from ..config import config
from ..utils import makedirs
import ast
import os
import threading
import platform
import re
from subprocess import Popen, PIPE
try:
    import Queue
except ImportError:
    import queue as Queue
if not hasattr(Queue.Queue, 'clear'):
    def _queue_clear(self):
        with self.mutex:
            self.queue.clear()
    setattr(Queue.Queue, 'clear', _queue_clear)

from .build_path import get_virtual_dirname, change_exts, add_path
from .build_dependant import get_include_file
from .build_flags import Flags
from .build_hash import path_hash, get_file_hash
from .build_latest_code import file_is_changed, code_need_to_rebuild, save_latest_state

OS_NAME = platform.system()
OS_IS_WINDOWS = OS_NAME == 'Windows'
OS_IS_LINUX = OS_NAME in ['Linux', 'Darwin']
assert OS_IS_WINDOWS or OS_IS_LINUX,\
    Exception('Unsupported Operator System: {}'.format(OS_NAME))

INC_PATHS = ['./']

# Load Config File
ENV_PATH = os.path.join(os.path.dirname(__file__), '../')
if os.path.dirname(config.BUILD_PATH) == '.':
    config.BUILD_PATH = os.path.join(ENV_PATH, config.BUILD_PATH)


def mkdir(dir_name):
    if not os.path.exists(dir_name):
        if config.SHOW_BUILDING_COMMAND:
            print('mkdir -p %s' % dir_name)
        makedirs(dir_name, exist_ok=True)


if OS_IS_LINUX:
    _rmdir_command = 'rm -rf'
elif OS_IS_WINDOWS:
    _rmdir_command = 'rd /s /q'


def rmdir(dir_name):
    # we use shell command to remove the non-empty or empry directory
    if os.path.exists(dir_name):
        command = '%s %s' % (_rmdir_command, dir_name)
        run_command(command)


def run_command(command):
    if config.SHOW_BUILDING_COMMAND:
        print(command)
    return os.system(command)


def run_command_parallel(commands, allow_error=False):
    command_queue = Queue.Queue()
    info_queue = Queue.Queue()
    for c in commands:
        command_queue.put(c)
    max_worker_num = min(config.MAX_BUILDING_WORKER_NUM, len(commands))
    for _ in range(max_worker_num):
        command_queue.put(None)

    def worker(command_queue, info_queue):
        while not command_queue.empty():
            e = command_queue.get()
            if e is None:
                break
            rtn = run_command(e)
            if rtn != 0:
                # Error
                command_queue.clear()
                info_queue.put(Exception('Error, terminated :-('))
    workers = [threading.Thread(target=worker, args=(command_queue, info_queue))
               for _ in range(max_worker_num)]
    for w in workers:
        w.daemon = True
    for w in workers:
        w.start()
    for w in workers:
        w.join()
    while not info_queue.empty():
        info = info_queue.get()
        if isinstance(info, Exception) and not allow_error:
            raise RuntimeError(info)


def command_exists(command):
    try:
        Popen([command], stdout=PIPE, stderr=PIPE, stdin=PIPE)
    except Exception:
        return False
    return True


class build_context:
    def __enter__(self):
        pass

    def __exit__(self, *dummy):
        save_latest_state()



"""Building Implementation"""
import os
import multiprocessing
try:
    from .build_utils import *
except Exception:
    from build_utils import *


NUM_CPU_CORE = multiprocessing.cpu_count()


def source_to_o(build_path, src_obj, compiler, cflags):
    mkdir(build_path)
    existed_dirs = set()
    updated = False
    commands = []
    for src, obj in src_obj:
        dir_name, obj_name = os.path.split(obj)
        build_dir_name = os.path.join(build_path, dir_name)
        build_name = os.path.join(build_path, dir_name, obj_name)
        if not code_need_to_rebuild(src) and os.path.exists(build_name):
            continue
        updated = True
        if build_dir_name not in existed_dirs:
            mkdir(build_dir_name)
            existed_dirs.add(build_dir_name)
        if OS_IS_WINDOWS and not command_exists(compiler):
            inc_flags = Flags()
            for path in INC_PATHS:
                p = os.path.join(ENV_PATH, path)
                inc_flags.add_string('-I{}'.format(p))
            cflags_sp = str(cflags).split()
            def_flags = ' '.join(
                [s for s in cflags_sp if len(s) > 2 and s[:2] == '-D'])
            command = 'cl /EHsc /O2 %s %s -c %s -Fo%s' % (
                def_flags, inc_flags, src, build_name)
        else:
            command = '%s %s %s -c -o %s' % (compiler, src, cflags, build_name)
        commands.append(command)
    run_command_parallel(commands)
    return updated


def o_to_so(target_name, objs, linker, ldflags):
    if OS_IS_WINDOWS and not command_exists(linker):
        command = 'link -DLL %s -out:%s' % (' '.join(objs), target_name)
    else:
        command = '%s %s %s -o %s' % (linker,
                                      ' '.join(objs), ldflags, target_name)
    run_command(command)


def source_to_so(build_path, srcs, target_name, compiler, cflags, ldflags, buildin_o=None):
    objs = change_exts(srcs, [('cpp', 'o')])
    if source_to_o(build_path, zip(srcs, objs), compiler, cflags) or\
            not os.path.exists(target_name):
        if buildin_o is not None:
            objs.extend(buildin_o)
        abs_objs = add_path(build_path, objs)
        o_to_so(target_name, abs_objs, compiler, ldflags)


def get_common_flags():
    HOST_NUM_THREADS = config.HOST_NUM_THREADS if config.HOST_NUM_THREADS > 0 else NUM_CPU_CORE
    COMMON_FLAGS = Flags().add_definition('HOST_NUM_THREADS', HOST_NUM_THREADS)
    if config.USING_OPTIMIZATION:
        COMMON_FLAGS.add_string('-O3')
    if config.DEBUG:
        COMMON_FLAGS.add_string('-g')
    COMMON_FLAGS.add_definition('USING_CBLAS', config.USING_CBLAS)
    INC_PATHS.extend(['./cpp/include'])
    for path in INC_PATHS:
        p = os.path.join(ENV_PATH, path)
        if p:
            COMMON_FLAGS.add_string('-I{}'.format(p))

    return COMMON_FLAGS


def get_build_flag_cpu():
    COMMON_FLAGS = get_common_flags()
    CFLAGS = Flags('-std=c++11').add_definition('USING_CUDA', 0).add_definition('USING_OPENMP', config.USING_OPENMP).\
        add_string(COMMON_FLAGS)
    if not OS_IS_WINDOWS:
        CFLAGS.add_string('-fPIC')
    LDFLAGS = Flags('-lpthread -shared')
    if config.USING_CBLAS:
        LDFLAGS.add_string('-lopenblas')
    if config.USING_OPENMP:
        CFLAGS.add_string('-fopenmp')
        LDFLAGS.add_string('-fopenmp')
    if config.USING_HIGH_LEVEL_WARNINGS:
        CFLAGS.add_string('-Werror -Wall -Wextra -pedantic -Wcast-align -Wcast-qual \
    -Wctor-dtor-privacy -Wdisabled-optimization -Wformat=2 -Winit-self -Wmissing-include-dirs \
    -Wold-style-cast -Woverloaded-virtual -Wredundant-decls -Wshadow \
    -Wsign-promo -Wundef -fdiagnostics-show-option')

    return config.CXX, CFLAGS, LDFLAGS


def get_build_flag_cuda():
    COMMON_FLAGS = get_common_flags()
    CU_FLAGS = Flags('-std=c++11 -x cu -Wno-deprecated-gpu-targets -dc \
    -D_MWAITXINTRIN_H_INCLUDED -D_FORCE_INLINES --expt-extended-lambda').\
        add_definition('USING_CUDA', 1).\
        add_definition('USING_HIP', 0).\
        add_string(COMMON_FLAGS)
    if not OS_IS_WINDOWS:
        CU_FLAGS.add_string('--compiler-options "-fPIC"')
    CU_LDFLAGS = Flags('-shared -Wno-deprecated-gpu-targets \
    -L%s/lib64 -lcuda -lcudart' % config.CUDA_DIR)
    if config.USING_CBLAS:
        CU_LDFLAGS.add_string('-lcublas')
    return config.NVCC, CU_FLAGS, CU_LDFLAGS


BUILD_FLAGS = dict(
    cpu=get_build_flag_cpu,
    cuda=get_build_flag_cuda,
)


def get_build_flag(ctx_name):
    flags = BUILD_FLAGS.get(ctx_name, None)()
    assert flags is not None, ValueError(
        'Unsupported Context: {} -('.format(ctx_name))
    return flags


def source_to_so_ctx(build_path, srcs, target_name, ctx_name):
    flags = get_build_flag(ctx_name)
    compiler, cflags, ldflags = flags[:3]

    buildin_path = os.path.join(
        config.BUILD_PATH, 'build', 'buildin', ctx_name)
    os.makedirs(buildin_path, exist_ok=True)

    buildin_o = []
    buildin_cpp = []
    for src in ['defines.cpp']:
        fname = os.path.join('cpp', 'src', src)
        buildin_o.append(os.path.join(buildin_path, fname))
        buildin_cpp.append(os.path.join(ENV_PATH, fname))
    buildin_o = change_exts(buildin_o, [('cpp', 'o')])

    for fname in buildin_o:
        if not os.path.exists(fname):
            with build_context():
                source_to_o('.', zip(
                    buildin_cpp, buildin_o), compiler, cflags)
    flags += (buildin_o, )
    source_to_so('.', srcs, target_name, *flags)



# Copyright (c) OpenMMLab. All rights reserved.
import math

import torch
from antgo.framework.helper.runner import get_dist_info
from torch.utils.data import Sampler
from antgo.framework.helper.utils.dist_utils import *

class ClassAwareSampler(Sampler):
    r"""Sampler that restricts data loading to the label of the dataset.

    A class-aware sampling strategy to effectively tackle the
    non-uniform class distribution. The length of the training data is
    consistent with source data. Simple improvements based on `Relay
    Backpropagation for Effective Learning of Deep Convolutional
    Neural Networks <https://arxiv.org/abs/1512.05830>`_

    The implementation logic is referred to
    https://github.com/Sense-X/TSD/blob/master/mmdet/datasets/samplers/distributed_classaware_sampler.py

    Args:
        dataset: Dataset used for sampling.
        samples_per_gpu (int): When model is :obj:`DistributedDataParallel`,
            it is the number of training samples on each GPU.
            When model is :obj:`DataParallel`, it is
            `num_gpus * samples_per_gpu`.
            Default : 1.
        num_replicas (optional): Number of processes participating in
            distributed training.
        rank (optional): Rank of the current process within num_replicas.
        seed (int, optional): random seed used to shuffle the sampler if
            ``shuffle=True``. This number should be identical across all
            processes in the distributed group. Default: 0.
        num_sample_class (int): The number of samples taken from each
            per-label list. Default: 1
    """

    def __init__(self,
                 dataset,
                 samples_per_gpu=1,
                 num_replicas=None,
                 rank=None,
                 seed=0,
                 num_sample_class=1):
        _rank, _num_replicas = get_dist_info()
        if num_replicas is None:
            num_replicas = _num_replicas
        if rank is None:
            rank = _rank

        self.dataset = dataset
        self.num_replicas = num_replicas
        self.samples_per_gpu = samples_per_gpu
        self.rank = rank
        self.epoch = 0
        # Must be the same across all workers. If None, will use a
        # random seed shared among workers
        # (require synchronization among all workers)
        self.seed = sync_random_seed(seed)

        # The number of samples taken from each per-label list
        assert num_sample_class > 0 and isinstance(num_sample_class, int)
        self.num_sample_class = num_sample_class
        # Get per-label image list from dataset
        assert hasattr(dataset, 'get_cat2imgs'), \
            'dataset must have `get_cat2imgs` function'
        self.cat_dict = dataset.get_cat2imgs()

        self.num_samples = int(
            math.ceil(
                len(self.dataset) * 1.0 / self.num_replicas /
                self.samples_per_gpu)) * self.samples_per_gpu
        self.total_size = self.num_samples * self.num_replicas

        # get number of images containing each category
        self.num_cat_imgs = [len(x) for x in self.cat_dict.values()]
        # filter labels without images
        self.valid_cat_inds = [
            i for i, length in enumerate(self.num_cat_imgs) if length != 0
        ]
        self.num_classes = len(self.valid_cat_inds)

    def __iter__(self):
        # deterministically shuffle based on epoch
        g = torch.Generator()
        g.manual_seed(self.epoch + self.seed)

        # initialize label list
        label_iter_list = RandomCycleIter(self.valid_cat_inds, generator=g)
        # initialize each per-label image list
        data_iter_dict = dict()
        for i in self.valid_cat_inds:
            data_iter_dict[i] = RandomCycleIter(self.cat_dict[i], generator=g)

        def gen_cat_img_inds(cls_list, data_dict, num_sample_cls):
            """Traverse the categories and extract `num_sample_cls` image
            indexes of the corresponding categories one by one."""
            id_indices = []
            for _ in range(len(cls_list)):
                cls_idx = next(cls_list)
                for _ in range(num_sample_cls):
                    id = next(data_dict[cls_idx])
                    id_indices.append(id)
            return id_indices

        # deterministically shuffle based on epoch
        num_bins = int(
            math.ceil(self.total_size * 1.0 / self.num_classes /
                      self.num_sample_class))
        indices = []
        for i in range(num_bins):
            indices += gen_cat_img_inds(label_iter_list, data_iter_dict,
                                        self.num_sample_class)

        # fix extra samples to make it evenly divisible
        if len(indices) >= self.total_size:
            indices = indices[:self.total_size]
        else:
            indices += indices[:(self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # subsample
        offset = self.num_samples * self.rank
        indices = indices[offset:offset + self.num_samples]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        self.epoch = epoch


class RandomCycleIter:
    """Shuffle the list and do it again after the list have traversed.

    The implementation logic is referred to
    https://github.com/wutong16/DistributionBalancedLoss/blob/master/mllt/datasets/loader/sampler.py

    Example:
        >>> label_list = [0, 1, 2, 4, 5]
        >>> g = torch.Generator()
        >>> g.manual_seed(0)
        >>> label_iter_list = RandomCycleIter(label_list, generator=g)
        >>> index = next(label_iter_list)
    Args:
        data (list or ndarray): The data that needs to be shuffled.
        generator: An torch.Generator object, which is used in setting the seed
            for generating random numbers.
    """  # noqa: W605

    def __init__(self, data, generator=None):
        self.data = data
        self.length = len(data)
        self.index = torch.randperm(self.length, generator=generator).numpy()
        self.i = 0
        self.generator = generator

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.data)

    def __next__(self):
        if self.i == self.length:
            self.index = torch.randperm(
                self.length, generator=self.generator).numpy()
            self.i = 0
        idx = self.data[self.index[self.i]]
        self.i += 1
        return idx



# -*- coding: UTF-8 -*-
# @Time    : 2018/11/9 5:50 PM
# @File    : random_dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.dataflow.dataset.dataset import Dataset
from antgo.utils import logger
import os
import copy
import sys
import time

class RandomDataset(Dataset):
  def __init__(self, train_or_test, dir, params={}):
    super(RandomDataset, self).__init__(train_or_test, dir, params)
    self._data_func = None
    logger.info('initialize random dataset')

  @property
  def size(self):
    return 10000

  def at(self, id):
    raise NotImplementedError

  def data_pool(self):
    assert(self._data_func is not None)

    for _ in range(self.size):
      data = self._data_func()
      yield data

  @property
  def data_func(self):
    return self._data_func

  @data_func.setter
  def data_func(self, val):
    self._data_func = val




from copy import deepcopy
import logging

import torch
import torch.nn as nn
import torch.nn.functional as F

class ModelEMA(nn.Module):
    def __init__(self, model, decay=0.9999, device=None):
        super().__init__()
        self.module = deepcopy(model)
        self.module.eval()
        self.decay = decay
        self.device = device
        if self.device is not None:
            self.module.to(device=device)

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)

    def _update(self, model, update_fn):
        with torch.no_grad():
            for ema_v, model_v in zip(self.module.parameters(), model.parameters()):
                if self.device is not None:
                    model_v = model_v.to(device=self.device)
                ema_v.copy_(update_fn(ema_v, model_v))
            for ema_v, model_v in zip(self.module.buffers(), model.buffers()):
                if self.device is not None:
                    model_v = model_v.to(device=self.device)
                ema_v.copy_(model_v)

    def update_parameters(self, model):
        self._update(model, update_fn=lambda e, m: self.decay * e + (1. - self.decay) * m)

    def state_dict(self):
        return self.module.state_dict()

    def load_state_dict(self, state_dict):
        self.module.load_state_dict(state_dict)




# -*- coding: UTF-8 -*-
# @Time    : 2019-06-17 17:26
# @File    : face_task.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.measures.base import *
from antgo.task.task import *

default = {'AntFaceVROC': ('FaceVROC', 'FACE_VERIFICATION')}

class AntFaceVROC(AntMeasure):
  def __init__(self, task):
    super(AntFaceVROC, self).__init__(task, "FaceVROC")
    assert(task.task_type=='FACE_VERIFICATION')
    self.dist_type = getattr(self.task, 'dist_type', 'l2')

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    person_db = {}
    same_person = []
    diff_person = []

    for det_person, person_annotation in data:
      person_pair_id = person_annotation['pair_id']
      person_issame = person_annotation['same']
      det_person_id = person_annotation['id']

      if person_issame and (person_pair_id not in person_db and det_person_id not in person_db):
        same_person.append((det_person_id, person_pair_id))
      elif person_pair_id not in person_db and det_person_id not in person_db:
        diff_person.append((det_person_id, person_pair_id))

      person_db[det_person_id] = det_person

    dist_list = []
    label_list = []
    for person_a_id, person_b_id in same_person:
      person_a_feat = person_db[person_a_id]
      person_b_feat = person_db[person_b_id]

      if self.dist_type == 'l2':
        dist_list.append(np.sqrt(np.sum(np.power(person_a_feat-person_b_feat, 2.0))))
      else:
        dist_list.append(-np.dot(person_a_feat, person_b_feat))

      label_list.append(1)

    for person_a_id, person_b_id in diff_person:
      person_a_feat = person_db[person_a_id]
      person_b_feat = person_db[person_b_id]

      if self.dist_type == 'l2':
        dist_list.append(np.sqrt(np.sum(np.power(person_a_feat-person_b_feat, 2.0))))
      else:
        dist_list.append(-np.dot(person_a_feat, person_b_feat))

      label_list.append(0)

    actual = np.array(label_list).astype(np.int64)
    posterior = np.array(dist_list)
    num = actual.shape[0]
    positive_num = len(np.where(actual == 1)[0])
    negative_num = num - positive_num

    true_positive_num = 0.0
    false_positive_num = 0.0

    sorted_x = sorted(zip(posterior, range(len(posterior))))

    fpr = []
    tpr = []
    for i in range(len(sorted_x)):
        if actual[sorted_x[i][1]] == 1:
            true_positive_num += 1.0
        else:
            false_positive_num += 1.0

        # false positive rate
        negative_num = negative_num if negative_num > 0 else negative_num + 0.0000000001
        fp_rate = false_positive_num / float(negative_num)
        # true positive rate
        positive_num = positive_num if positive_num > 0 else positive_num + 0.0000000001
        tp_rate = true_positive_num / float(positive_num)

        fpr.append(fp_rate)
        tpr.append(tp_rate)

    trp_0dot01 = np.interp(0.01, fpr, tpr)
    trp_0dot001 = np.interp(0.001, fpr, tpr)

    return {'statistic': {'name': self.name,
                          'value': [{'name': 'TRP@0.01',
                                     'value': trp_0dot01,
                                     'type': 'SCALAR',
                                     'x': 'x',
                                     'y': 'y'},
                                    {'name': 'TRP@0.001',
                                      'value': trp_0dot001,
                                      'type': 'SCALAR',
                                      'x': 'x',
                                      'y': 'y'},
                                    {'name': 'ROC',
                                     'value': np.array([fpr,tpr]).transpose(),
                                     'type': 'CURVE',
                                     'x': 'FPR',
                                     'y': 'TRP'}]}}


class AntFaceIRank(AntMeasure):
  def __init__(self, task):
    super(AntFaceIRank, self).__init__(task, "FaceIRank")
    assert(task.task_type=='FACE_IDENTIFICATION')

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/24 23:11
# @File    : inpainting.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import torch
import mmcv
import torch
from mmcv.parallel import collate, scatter
from mmedit.apis import init_model, inpainting_inference
from mmedit.core import tensor2img
from mmedit.datasets.pipelines import Compose
import numpy as np


def _inpainting_inference(model, masked_img, mask):
  """Inference image with the model.

  Args:
      model (nn.Module): The loaded model.

  Returns:
      Tensor: The predicted inpainting result.
  """
  device = next(model.parameters()).device  # model device

  infer_pipeline = [
    # dict(type='LoadImageFromFile', key='masked_img'),
    # dict(type='LoadMask', mask_mode='file', mask_config=dict()),
    dict(type='Pad', keys=['masked_img', 'mask'], mode='reflect'),
    dict(
      type='Normalize',
      keys=['masked_img'],
      mean=[127.5] * 3,
      std=[127.5] * 3,
      to_rgb=False),
    dict(type='GetMaskedImage', img_name='masked_img'),
    dict(
      type='Collect',
      keys=['masked_img', 'mask'],
      meta_keys=[]),
    dict(type='ImageToTensor', keys=['masked_img', 'mask'])
  ]

  # build the data pipeline
  test_pipeline = Compose(infer_pipeline)
  # prepare data
  data = dict(masked_img=masked_img, mask=mask)
  data = test_pipeline(data)
  data = collate([data], samples_per_gpu=1)
  if 'cuda' in str(device):
    data = scatter(data, [device])[0]
  else:
    data.pop('meta')
  # forward the model
  with torch.no_grad():
    result = model(test_mode=True, **data)

  return result['fake_img']


class Inpainting(object):
  def __init__(self, config_file, checkpoint_file, device='cpu'):
    self.model = init_model(config_file, checkpoint_file, device=device)

  def __call__(self, *args, **kwargs):
    masked_img = None
    mask = None
    if len(args) == 2:
      masked_img, mask = args
      if len(mask.shape) == 2:
        mask = np.expand_dims(mask, -1)
      else:
        mask = mask[:,:,:1]
    else:
      masked_img = args[0]
      h,w = masked_img.shape[:2]
      mask = np.zeros((h,w,1))

    result = _inpainting_inference(self.model, masked_img, mask)
    result = tensor2img(result, min_max=(-1, 1))[..., ::-1]

    return result['fake_img']


# Copyright (c) OpenMMLab. All rights reserved.
from .dist_utils import (DistOptimizerHook, all_reduce_dict, allreduce_grads,
                         reduce_mean)
from .misc import (center_of_mass, filter_scores_and_topk, flip_tensor,
                   generate_coordinate, mask2ndarray, multi_apply,
                   select_single_mlvl, unmap)

__all__ = [
    'allreduce_grads', 'DistOptimizerHook', 'reduce_mean', 'multi_apply',
    'unmap', 'mask2ndarray', 'flip_tensor', 'all_reduce_dict',
    'center_of_mass', 'generate_coordinate', 'select_single_mlvl',
    'filter_scores_and_topk'
]



from .ssh_submit import *
from .local_submit import *
from .custom_submit import *
__all__=[
    'ssh_submit_process_func', 
    'ssh_submit_resource_check_func', 
    'local_submit_process_func', 
    'local_submit_resource_check_func',
    'custom_submit_process_func',
    'custom_submit_resource_check_func'
]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# Copyright 2019 Harrison Erd
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import sys
import os
import signal
import pickle
from threading import Thread


def load(location, auto_dump, sig=True):
    '''Return a pickledb object. location is the path to the json file.'''
    return PickleDB(location, auto_dump, sig)


class PickleDB(object):

    key_string_error = TypeError('Key/name must be a string!')

    def __init__(self, location, auto_dump, sig):
        '''Creates a database object and loads the data from the location path.
        If the file does not exist it will be created on the first update.
        '''
        self.load(location, auto_dump)
        self.dthread = None
        if sig:
            self.set_sigterm_handler()

    def __getitem__(self, item):
        '''Syntax sugar for get()'''
        return self.get(item)

    def __setitem__(self, key, value):
        '''Sytax sugar for set()'''
        return self.set(key, value)

    def __delitem__(self, key):
        '''Sytax sugar for rem()'''
        return self.rem(key)

    def set_sigterm_handler(self):
        '''Assigns sigterm_handler for graceful shutdown during dump()'''
        def sigterm_handler():
            if self.dthread is not None:
                self.dthread.join()
            sys.exit(0)
        signal.signal(signal.SIGTERM, sigterm_handler)

    def load(self, location, auto_dump):
        '''Loads, reloads or changes the path to the db file'''
        location = os.path.expanduser(location)
        self.loco = location
        self.auto_dump = auto_dump
        if os.path.exists(location):
            self._loaddb()
        else:
            self.db = {}
        return True

    def dump(self):
        '''Force dump memory db to file'''
        pickle.dump(self.db, open(self.loco, 'wb'))
        # self.dthread = Thread(
        #     target=json.dump,
        #     args=(self.db, open(self.loco, 'wt')))
        # self.dthread.start()
        # self.dthread.join()
        return True

    def _loaddb(self):
        '''Load or reload the json info from the file'''
        try:
            # self.db = json.load(open(self.loco, 'rt'))
            self.db = pickle.load(open(self.loco, 'rb'))
        except ValueError:
            if os.stat(self.loco).st_size == 0:  # Error raised because file is empty
                self.db = {}
            else:
                raise  # File is not empty, avoid overwriting it

    def _autodumpdb(self):
        '''Write/save the json dump into the file if auto_dump is enabled'''
        if self.auto_dump:
            self.dump()

    def set(self, key, value):
        '''Set the str value of a key'''
        if isinstance(key, str):
            self.db[key] = value
            self._autodumpdb()
            return True
        else:
            raise self.key_string_error

    def get(self, key):
        '''Get the value of a key'''
        try:
            return self.db[key]
        except KeyError:
            return False

    def getall(self):
        '''Return a list of all keys in db'''
        return self.db.keys()

    def exists(self, key):
        '''Return True if key exists in db, return False if not'''
        return key in self.db

    def rem(self, key):
        '''Delete a key'''
        if not key in self.db: # return False instead of an exception
            return False
        del self.db[key]
        self._autodumpdb()
        return True

    def totalkeys(self, name=None):
        '''Get a total number of keys, lists, and dicts inside the db'''
        if name is None:
            total = len(self.db)
            return total
        else:
            total = len(self.db[name])
            return total

    def append(self, key, more):
        '''Add more to a key's value'''
        tmp = self.db[key]
        self.db[key] = tmp + more
        self._autodumpdb()
        return True

    def lcreate(self, name):
        '''Create a list, name must be str'''
        if isinstance(name, str):
            self.db[name] = []
            self._autodumpdb()
            return True
        else:
            raise self.key_string_error

    def ladd(self, name, value):
        '''Add a value to a list'''
        self.db[name].append(value)
        self._autodumpdb()
        return True

    def lextend(self, name, seq):
        '''Extend a list with a sequence'''
        self.db[name].extend(seq)
        self._autodumpdb()
        return True

    def lgetall(self, name):
        '''Return all values in a list'''
        return self.db[name]

    def lget(self, name, pos):
        '''Return one value in a list'''
        return self.db[name][pos]

    def lrange(self, name, start=None, end=None):
        '''Return range of values in a list '''
        return self.db[name][start:end]

    def lremlist(self, name):
        '''Remove a list and all of its values'''
        number = len(self.db[name])
        del self.db[name]
        self._autodumpdb()
        return number

    def lremvalue(self, name, value):
        '''Remove a value from a certain list'''
        self.db[name].remove(value)
        self._autodumpdb()
        return True

    def lpop(self, name, pos):
        '''Remove one value in a list'''
        value = self.db[name][pos]
        del self.db[name][pos]
        self._autodumpdb()
        return value

    def llen(self, name):
        '''Returns the length of the list'''
        return len(self.db[name])

    def lappend(self, name, pos, more):
        '''Add more to a value in a list'''
        tmp = self.db[name][pos]
        self.db[name][pos] = tmp + more
        self._autodumpdb()
        return True

    def lexists(self, name, value):
        '''Determine if a value  exists in a list'''
        return value in self.db[name]

    def dcreate(self, name):
        '''Create a dict, name must be str'''
        if isinstance(name, str):
            self.db[name] = {}
            self._autodumpdb()
            return True
        else:
            raise self.key_string_error

    def dadd(self, name, pair):
        '''Add a key-value pair to a dict, "pair" is a tuple'''
        self.db[name][pair[0]] = pair[1]
        self._autodumpdb()
        return True

    def dget(self, name, key):
        '''Return the value for a key in a dict'''
        return self.db[name][key]

    def dgetall(self, name):
        '''Return all key-value pairs from a dict'''
        return self.db[name]

    def drem(self, name):
        '''Remove a dict and all of its pairs'''
        del self.db[name]
        self._autodumpdb()
        return True

    def dpop(self, name, key):
        '''Remove one key-value pair in a dict'''
        value = self.db[name][key]
        del self.db[name][key]
        self._autodumpdb()
        return value

    def dkeys(self, name):
        '''Return all the keys for a dict'''
        return self.db[name].keys()

    def dvals(self, name):
        '''Return all the values for a dict'''
        return self.db[name].values()

    def dexists(self, name, key):
        '''Determine if a key exists or not in a dict'''
        return key in self.db[name]

    def dmerge(self, name1, name2):
        '''Merge two dicts together into name1'''
        first = self.db[name1]
        second = self.db[name2]
        first.update(second)
        self._autodumpdb()
        return True

    def deldb(self):
        '''Delete everything from the database'''
        self.db = {}
        self._autodumpdb()
        return True



# -*- coding: UTF-8 -*-
# @Time    : 2022/5/1 15:03
# @File    : data.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.framework.torch2paddle.mapper.io import *


from ..runner.builder import *
import copy
import json
import numpy as np
import time


@MEASURES.register_module()
class OKS(object):
    def __init__(self, sigma=0.1) -> None:
        self.sigma = sigma

    def keys(self):
        # 约束使用此评估方法，需要具体的关键字信息
        return {'pred': ['pred_joints2d'], 'gt': ['joints2d', 'bboxes', 'joints_vis']}

    def __call__(self, preds, gts):
        oks_list = []
        for pred, gt in zip(preds,gts):
            # 预测信息
            pred_joints2d = pred['pred_joints2d']   # Nx2

            # GT信息
            gt_joints2d = gt['joints2d']            # Nx2            
            gt_bboxes = gt['bboxes']                # 4
            gt_joints_vis = gt['joints_vis']        # N

            if np.sum(gt_joints_vis) == 0:
                continue

            gt_joints_vis = np.array(gt_joints_vis).reshape(-1, 1)
            x0,y0,x1,y1 = gt_bboxes            
            joints_2d_dist = np.sum(np.power(pred_joints2d[:,:2]-gt_joints2d, 2.0), -1,keepdims=True)
            Z = 2.0 * (x1-x0)*(y1-y0) * self.sigma

            oks_v = np.sum(np.exp(-joints_2d_dist/Z)*gt_joints_vis)/np.sum(gt_joints_vis)
            oks_list.append(oks_v)

        oks_err = float(np.mean(oks_list))
        print()
        print("======== OKS Acc: %f ========" % (oks_err))

        error_info = {
            'oks': oks_err
        }
        return error_info



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn.functional as F
from antgo.framework.helper.cnn import ConvModule
from antgo.framework.helper.base_module import BaseModule, ModuleList


class ConvUpsample(BaseModule):
    """ConvUpsample performs 2x upsampling after Conv.

    There are several `ConvModule` layers. In the first few layers, upsampling
    will be applied after each layer of convolution. The number of upsampling
    must be no more than the number of ConvModule layers.

    Args:
        in_channels (int): Number of channels in the input feature map.
        inner_channels (int): Number of channels produced by the convolution.
        num_layers (int): Number of convolution layers.
        num_upsample (int | optional): Number of upsampling layer. Must be no
            more than num_layers. Upsampling will be applied after the first
            ``num_upsample`` layers of convolution. Default: ``num_layers``.
        conv_cfg (dict): Config dict for convolution layer. Default: None,
            which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer. Default: None.
        init_cfg (dict): Config dict for initialization. Default: None.
        kwargs (key word augments): Other augments used in ConvModule.
    """

    def __init__(self,
                 in_channels,
                 inner_channels,
                 num_layers=1,
                 num_upsample=None,
                 conv_cfg=None,
                 norm_cfg=None,
                 init_cfg=None,
                 **kwargs):
        super(ConvUpsample, self).__init__(init_cfg)
        if num_upsample is None:
            num_upsample = num_layers
        assert num_upsample <= num_layers, \
            f'num_upsample({num_upsample})must be no more than ' \
            f'num_layers({num_layers})'
        self.num_layers = num_layers
        self.num_upsample = num_upsample
        self.conv = ModuleList()
        for i in range(num_layers):
            self.conv.append(
                ConvModule(
                    in_channels,
                    inner_channels,
                    3,
                    padding=1,
                    stride=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    **kwargs))
            in_channels = inner_channels

    def forward(self, x):
        num_upsample = self.num_upsample
        for i in range(self.num_layers):
            x = self.conv[i](x)
            if num_upsample > 0:
                num_upsample -= 1
                x = F.interpolate(
                    x, scale_factor=2, mode='bilinear', align_corners=False)
        return x



import warnings
from collections.abc import Sequence
from typing import List, Optional, Tuple, Union

import numpy as np
import torch
from antgo.framework.helper.models.detectors.core.mask.structures import BitmapMasks
from torch.nn import functional as F


def bbox2points(box):
    min_x, min_y, max_x, max_y = torch.split(box[:, :4], [1, 1, 1, 1], dim=1)

    return torch.cat(
        [min_x, min_y, max_x, min_y, max_x, max_y, min_x, max_y], dim=1
    ).reshape(
        -1, 2
    )  # n*4,2


def points2bbox(point, max_w, max_h):
    point = point.reshape(-1, 4, 2)
    if point.size()[0] > 0:
        min_xy = point.min(dim=1)[0]
        max_xy = point.max(dim=1)[0]
        xmin = min_xy[:, 0].clamp(min=0, max=max_w)
        ymin = min_xy[:, 1].clamp(min=0, max=max_h)
        xmax = max_xy[:, 0].clamp(min=0, max=max_w)
        ymax = max_xy[:, 1].clamp(min=0, max=max_h)
        min_xy = torch.stack([xmin, ymin], dim=1)
        max_xy = torch.stack([xmax, ymax], dim=1)
        return torch.cat([min_xy, max_xy], dim=1)  # n,4
    else:
        return point.new_zeros(0, 4)


def check_is_tensor(obj):
    """Checks whether the supplied object is a tensor."""
    if not isinstance(obj, torch.Tensor):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(type(obj)))


def normal_transform_pixel(
    height: int,
    width: int,
    eps: float = 1e-14,
    device: Optional[torch.device] = None,
    dtype: Optional[torch.dtype] = None,
) -> torch.Tensor:
    tr_mat = torch.tensor(
        [[1.0, 0.0, -1.0], [0.0, 1.0, -1.0], [0.0, 0.0, 1.0]],
        device=device,
        dtype=dtype,
    )  # 3x3

    # prevent divide by zero bugs
    width_denom: float = eps if width == 1 else width - 1.0
    height_denom: float = eps if height == 1 else height - 1.0

    tr_mat[0, 0] = tr_mat[0, 0] * 2.0 / width_denom
    tr_mat[1, 1] = tr_mat[1, 1] * 2.0 / height_denom

    return tr_mat.unsqueeze(0)  # 1x3x3


def normalize_homography(
    dst_pix_trans_src_pix: torch.Tensor,
    dsize_src: Tuple[int, int],
    dsize_dst: Tuple[int, int],
) -> torch.Tensor:
    check_is_tensor(dst_pix_trans_src_pix)

    if not (
        len(dst_pix_trans_src_pix.shape) == 3
        or dst_pix_trans_src_pix.shape[-2:] == (3, 3)
    ):
        raise ValueError(
            "Input dst_pix_trans_src_pix must be a Bx3x3 tensor. Got {}".format(
                dst_pix_trans_src_pix.shape
            )
        )

    # source and destination sizes
    src_h, src_w = dsize_src
    dst_h, dst_w = dsize_dst

    # compute the transformation pixel/norm for src/dst
    src_norm_trans_src_pix: torch.Tensor = normal_transform_pixel(src_h, src_w).to(
        dst_pix_trans_src_pix
    )
    src_pix_trans_src_norm = torch.inverse(src_norm_trans_src_pix.float()).to(
        src_norm_trans_src_pix.dtype
    )
    dst_norm_trans_dst_pix: torch.Tensor = normal_transform_pixel(dst_h, dst_w).to(
        dst_pix_trans_src_pix
    )

    # compute chain transformations
    dst_norm_trans_src_norm: torch.Tensor = dst_norm_trans_dst_pix @ (
        dst_pix_trans_src_pix @ src_pix_trans_src_norm
    )
    return dst_norm_trans_src_norm


def warp_affine(
    src: torch.Tensor,
    M: torch.Tensor,
    dsize: Tuple[int, int],
    mode: str = "bilinear",
    padding_mode: str = "zeros",
    align_corners: Optional[bool] = None,
) -> torch.Tensor:
    if not isinstance(src, torch.Tensor):
        raise TypeError(
            "Input src type is not a torch.Tensor. Got {}".format(type(src))
        )

    if not isinstance(M, torch.Tensor):
        raise TypeError("Input M type is not a torch.Tensor. Got {}".format(type(M)))

    if not len(src.shape) == 4:
        raise ValueError("Input src must be a BxCxHxW tensor. Got {}".format(src.shape))

    if not (len(M.shape) == 3 or M.shape[-2:] == (2, 3)):
        raise ValueError("Input M must be a Bx2x3 tensor. Got {}".format(M.shape))

    # TODO: remove the statement below in kornia v0.6
    if align_corners is None:
        message: str = (
            "The align_corners default value has been changed. By default now is set True "
            "in order to match cv2.warpAffine."
        )
        warnings.warn(message)
        # set default value for align corners
        align_corners = True

    B, C, H, W = src.size()

    # we generate a 3x3 transformation matrix from 2x3 affine

    dst_norm_trans_src_norm: torch.Tensor = normalize_homography(M, (H, W), dsize)

    src_norm_trans_dst_norm = torch.inverse(dst_norm_trans_src_norm.float())

    grid = F.affine_grid(
        src_norm_trans_dst_norm[:, :2, :],
        [B, C, dsize[0], dsize[1]],
        align_corners=align_corners,
    )

    return F.grid_sample(
        src.float(),
        grid,
        align_corners=align_corners,
        mode=mode,
        padding_mode=padding_mode,
    ).to(src.dtype)


class Transform2D:
    @staticmethod
    def transform_bboxes(bbox, M, out_shape):
        if isinstance(bbox, Sequence):
            assert len(bbox) == len(M)
            return [
                Transform2D.transform_bboxes(b, m, o)
                for b, m, o in zip(bbox, M, out_shape)
            ]
        else:
            if bbox.shape[0] == 0:
                return bbox
            score = None
            if bbox.shape[1] > 4:
                score = bbox[:, 4:]
            points = bbox2points(bbox[:, :4])
            points = torch.cat(
                [points, points.new_ones(points.shape[0], 1)], dim=1
            )  # n,3
            points = torch.matmul(M, points.t()).t()
            points = points[:, :2] / points[:, 2:3]
            bbox = points2bbox(points, out_shape[1], out_shape[0])
            if score is not None:
                return torch.cat([bbox, score], dim=1)
            return bbox

    @staticmethod
    def transform_masks(
        mask: Union[BitmapMasks, List[BitmapMasks]],
        M: Union[torch.Tensor, List[torch.Tensor]],
        out_shape: Union[list, List[list]],
    ):
        if isinstance(mask, Sequence):
            assert len(mask) == len(M)
            return [
                Transform2D.transform_masks(b, m, o)
                for b, m, o in zip(mask, M, out_shape)
            ]
        else:
            if mask.shape[0] == 0:
                return torch.empty((0, *out_shape), device=M.device)

            mask_tensor = mask[:, None, ...]
            return warp_affine(
                    mask_tensor,
                    M[None, ...].expand(mask.shape[0], -1, -1),
                    out_shape,
                ).squeeze(1)


    @staticmethod
    def transform_image(img, M, out_shape):
        if isinstance(img, Sequence):
            assert len(img) == len(M)
            return [
                Transform2D.transform_image(b, m, shape)
                for b, m, shape in zip(img, M, out_shape)
            ]
        else:
            if img.dim() == 2:
                img = img[None, None, ...]
            elif img.dim() == 3:
                img = img[None, ...]

            return (
                warp_affine(img.float(), M[None, ...], out_shape, mode="nearest")
                .squeeze()
                .to(img.dtype)
            )

def random_ignore(bbox, label=None, score=None, mask=None, around_bbox=None, index=None):
    if bbox.size(0) == 0:
        return bbox, label, mask, index, around_bbox

    valid = np.random.random([bbox.shape[0]]) > 0.5
    valid = torch.from_numpy(valid).to(bbox.device) 
    
    bbox = bbox[valid]
    if label is not None:
        label = label[valid]
    if mask is not None:
        mask = BitmapMasks(mask.masks[valid.cpu().numpy()], mask.height, mask.width)
    if index is not None:
        index = index[valid]
    if around_bbox is not None:
        around_bbox = around_bbox[valid]

    return bbox, label, mask, index, around_bbox


def filter_invalid(bbox, label=None, score=None, mask=None, around_bbox=None, index=None, thr=0.0, box_min_size=0, class_constraint=None, aspect_ratio=None):
    if score is not None:
        valid = score > thr
        bbox = bbox[valid]
        if label is not None:
            label = label[valid]
        if mask is not None:
            mask = BitmapMasks(mask.masks[valid.cpu().numpy()], mask.height, mask.width)
        if index is not None:
            index = index[valid]
        if around_bbox is not None:
            around_bbox = around_bbox[valid]

    if box_min_size is not None:
        bw = bbox[:, 2] - bbox[:, 0]
        bh = bbox[:, 3] - bbox[:, 1]
        valid = (bw > box_min_size) & (bh > box_min_size)
        bbox = bbox[valid]
        if label is not None:
            label = label[valid]
        if mask is not None:
            mask = BitmapMasks(mask.masks[valid.cpu().numpy()], mask.height, mask.width)
        if index is not None:
            index = index[valid]
        if around_bbox is not None:
            around_bbox = around_bbox[valid]
    
    if aspect_ratio is not None:
        bw = bbox[:, 2] - bbox[:, 0]
        bh = bbox[:, 3] - bbox[:, 1]
        b_aspect_ratio = bw/bh
        valid = (b_aspect_ratio >= aspect_ratio[0]) & (b_aspect_ratio <= aspect_ratio[1])
        bbox = bbox[valid]
        if label is not None:
            label = label[valid]
        if mask is not None:
            mask = BitmapMasks(mask.masks[valid.cpu().numpy()], mask.height, mask.width)
        if index is not None:
            index = index[valid]
        if around_bbox is not None:
            around_bbox = around_bbox[valid]

    if class_constraint is not None and label is not None and label.shape[0] > 0:
        # class_constraint {0: (min,max), 1: (min,max)}
        # 如果不满足约束条件，则全部数据无效
        valid = torch.ones_like(label).bool()
        for class_label, info in class_constraint.items():
            class_label_num = torch.sum((label == (int)(class_label)).int())
            if class_label_num < info['min'] or class_label_num > info['max']:
                valid = torch.zeros_like(label).bool()
                break
        
        bbox = bbox[valid]
        if label is not None:
            label = label[valid]
        if mask is not None:
            mask = BitmapMasks(mask.masks[valid.cpu().numpy()], mask.height, mask.width)
        if index is not None:
            index = index[valid]
        if around_bbox is not None:
            around_bbox = around_bbox[valid]            

    return bbox, label, mask, index, around_bbox


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path as osp
import sys


def get_dep_file_path(current_file_path, from_level, from_str):
    """ 根据from信息获取依赖包所在文件。如果from字符串中存在相对路径（出现"."），
        则根据相对路径找到相应的文件；反之，执行import语句找到相应依赖的文件。
        Args:
            current_file_path (str): from信息所在文件的路径。
            from_level (int): from信息中父目录级别数。
            from_str (str): from信息中依赖包名字。
    """
    if from_level > 0:
        while from_level > 0:
            current_file_path, folder_or_file = osp.split(current_file_path)
            from_level -= 1
        if from_str is None:
            import_file_path = osp.join(current_file_path, "__init__.py")
        else:
            current_file_path = osp.join(current_file_path,
                                         osp.join(*from_str.split(".")))
            if osp.exists(current_file_path + ".py"):
                import_file_path = current_file_path + ".py"
            else:
                import_file_path = osp.join(current_file_path, "__init__.py")
    else:
        current_abs_path = osp.dirname(current_file_path)
        sys.path.append(current_abs_path)
        if len(from_str.split(".")) == 1:
            key_str = from_str
            exec("import {}".format(key_str))
        else:
            from_seg = from_str.split(".")
            from_str = ".".join(from_seg[0:-1])
            key_str = from_seg[-1]
            exec("from {} import {}".format(from_str, key_str))
        sys.path.pop(-1)
        import_file_path = locals()[key_str].__file__
    return import_file_path


def add_line_continuation_symbol(code):
    code_list = code.split("\n")
    for i, line in enumerate(code_list):
        if line.strip().endswith("="):
            code_list[i] = line + "\\"
    return "\n".join(code_list)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import grad
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class Sobolev(nn.Module):
	'''
	Sobolev Training for Neural Networks
	https://arxiv.org/pdf/1706.04859.pdf
	Knowledge Transfer with Jacobian Matching
	http://de.arxiv.org/pdf/1803.00443
	'''
	def __init__(self):
		super(Sobolev, self).__init__()

	def forward(self, out_s, out_t, img, target):
		target_out_s = torch.gather(out_s, 1, target.view(-1, 1))
		grad_s       = grad(outputs=target_out_s, inputs=img,
							grad_outputs=torch.ones_like(target_out_s),
							create_graph=True, retain_graph=True, only_inputs=True)[0]
		norm_grad_s  = F.normalize(grad_s.view(grad_s.size(0), -1), p=2, dim=1)

		target_out_t = torch.gather(out_t, 1, target.view(-1, 1))
		grad_t       = grad(outputs=target_out_t, inputs=img,
							grad_outputs=torch.ones_like(target_out_t),
							create_graph=True, retain_graph=True, only_inputs=True)[0]
		norm_grad_t  = F.normalize(grad_t.view(grad_t.size(0), -1), p=2, dim=1)

		loss = F.mse_loss(norm_grad_s, norm_grad_t.detach())

		return loss


import cv2
import numpy as np
import torch
from pytorch3d.transforms.rotation_conversions import matrix_to_axis_angle


def vis_2d_points_img(show_image, joints):
    BoneVec_new = [
        (1, 2),
        (1, 5),
        (5, 9),
        (9, 13),
        (13, 17),
        (1, 17),
        (2, 3),
        (3, 4),
        (5, 6),
        (6, 7),
        (7, 8),
        (9, 10),
        (10, 11),
        (11, 12),
        (13, 14),
        (14, 15),
        (15, 16),
        (17, 18),
        (18, 19),
        (19, 20),
    ]

    BoneColorVec_point_21 = [
        (0, 0, 200),
        (0, 0, 200),
        (150, 0, 0),
        (200, 0, 0),
        (255, 0, 0),
        (100, 100, 0),
        (150, 150, 0),
        (200, 200, 0),
        (255, 255, 0),
        (0, 100, 50),
        (0, 150, 75),
        (0, 200, 100),
        (0, 255, 125),
        (0, 50, 100),
        (0, 75, 150),
        (0, 100, 200),
        (0, 125, 255),
        (120, 0, 120),
        (150, 0, 150),
        (200, 0, 200),
        (255, 0, 255),
    ]

    for jointIndex in range(len(BoneVec_new)):
        x1 = int(joints[BoneVec_new[jointIndex][0]][0])
        y1 = int(joints[BoneVec_new[jointIndex][0]][1])
        joint1 = (x1, y1)
        cv2.circle(show_image, joint1, 1, BoneColorVec_point_21[BoneVec_new[jointIndex][0]], -1)
        x2 = int(joints[BoneVec_new[jointIndex][1]][0])
        y2 = int(joints[BoneVec_new[jointIndex][1]][1])
        joint2 = (x2, y2)
        cv2.circle(show_image, joint2, 1, BoneColorVec_point_21[BoneVec_new[jointIndex][0]], -1)
        cv2.line(show_image, joint1, joint2, (0, 0, 0), 1)

    return show_image


def decode_keypoint(heatmap):
    kpys = []
    h, w = heatmap.shape[1:3]
    for i in range(21):
        hm = heatmap[i]
        idx = hm.argmax()
        x = idx % w
        y = idx // h
        kpys.append([x, y])
    return kpys


def draw_skeleton(image, kps, radius=1):
    # lines_index = [[0,13,14,15,16], [0,1,2,3,17], [0,4,5,6,18], [0,10,11,12,19], [0,7,8,9,20]]      # Interhand
    lines_index = [
        [0, 1, 2, 3, 4],
        [0, 5, 6, 7, 8],
        [0, 9, 10, 11, 12],
        [0, 13, 14, 15, 16],
        [0, 17, 18, 19, 20],
    ]  # predict monapth
    colors = [(255, 0, 0), (0, 255, 0), (0, 0, 255), (255, 255, 0), (255, 0, 255)]

    ori_pt = kps[0]
    x, y = int(ori_pt[0]), int(ori_pt[1])
    cv2.circle(image, (x, y), radius=radius + 1, color=colors[0], thickness=-1)

    for n, l_idx in enumerate(lines_index):
        for i in range(1, 5):
            x, y = int(kps[l_idx[i]][0]), int(kps[l_idx[i]][1])
            cv2.circle(image, (x, y), radius=radius + 1, color=colors[n], thickness=-1)

            x0, y0 = int(kps[l_idx[i - 1]][0]), int(kps[l_idx[i - 1]][1])
            cv2.line(image, (x0, y0), (x, y), colors[n], radius)

    return image


def check_img_to_save(img_path, img, label, joint, joint2):
    img_path = img_path.replace("/", "#")
    # if label is not None:
    #     after_kpts = decode_keypoint(label)
    after_kpts = label
    test_image = (img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
    test_image = cv2.cvtColor(test_image, cv2.COLOR_GRAY2RGB)
    test_image2 = test_image.copy()
    test_image3 = test_image.copy()    
    if label is not None:
        test_image = draw_skeleton(test_image, np.array(after_kpts) * 8)
        # test_image = draw_skeleton(test_image, np.array(label) * 8)
    test_image2 = draw_skeleton(test_image2, np.array(joint[:, :2]) * 8)
    test_image3 = draw_skeleton(test_image3, np.array(joint2[:, :2]) * 8)
    
    test_image = cv2.resize(test_image, (128, 128))
    test_image2 = cv2.resize(test_image2, (128, 128))
    test_image3 = cv2.resize(test_image3, (128, 128))
    saveimg = np.concatenate([test_image, test_image2, test_image3], axis=1)
    import os

    if not os.path.exists("outputs/check"):
        os.mkdir("outputs/check")
    cv2.imwrite("outputs/check/{}.jpg".format(img_path), saveimg)


def check_img_and_history_to_save(img_path, img, history_img, label, his_label):
    img_path = img_path.replace("/", "#")
    if label is not None:
        after_kpts = decode_keypoint(label)
    if his_label is not None:
        his_after_kpts = decode_keypoint(his_label)
    test_image = (img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
    test_image = cv2.cvtColor(test_image, cv2.COLOR_GRAY2RGB)
    test_image2 = (history_img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
    test_image2 = cv2.cvtColor(test_image2, cv2.COLOR_GRAY2RGB)
    if label is not None:
        test_image = draw_skeleton(test_image, np.array(after_kpts) * 6)
        # test_image = draw_skeleton(test_image, np.array(label) * 8)
    if his_label is not None:
        test_image2 = draw_skeleton(test_image2, np.array(his_after_kpts) * 6)
    # test_image2 = draw_skeleton(test_image2, np.array(joint[:, :2]) * 8)

    test_image = cv2.resize(test_image, (128, 128))
    test_image2 = cv2.resize(test_image2, (128, 128))
    saveimg = np.concatenate([test_image, test_image2], axis=1)
    cv2.imwrite("outputs/check/{}.jpg".format(img_path), saveimg)


def eval_img_to_save(img_path, img, pred_joint, joint):
    img = img.cpu()
    joint = joint.cpu()
    pred_joint = pred_joint.cpu()
    test_image = (img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
    test_image = cv2.cvtColor(test_image, cv2.COLOR_GRAY2RGB)
    test_image2 = test_image.copy()
    test_image = draw_skeleton(test_image, np.array(pred_joint[:, :2]) * 8)
    test_image2 = draw_skeleton(test_image2, np.array(joint[:, :2]) * 8)

    test_image = cv2.resize(test_image, (128, 128))
    test_image2 = cv2.resize(test_image2, (128, 128))
    saveimg = np.concatenate([test_image, test_image2], axis=1)
    cv2.imwrite("outputs/check_ori/{}.jpg".format(img_path), saveimg)


def check_dof_img_to_save(img_path, img, label, joint, pose, has_common_dof):
    import torch
    from manopth import demo
    from manopth.manolayer import ManoLayer
    from matplotlib import pyplot as plt

    img_path = img_path.replace("/", "#")

    if has_common_dof:
        mano_layer = ManoLayer(
            mano_root="/mnt/bn/wlxlmk/mlx/users/huhui.22/playground/manodatas/models",
            use_pca=False,
            flat_hand_mean=False,
        )

        pose = torch.tensor(pose.astype(np.float32))
        verts, joints = mano_layer.forward(pose.reshape(-1, 48), torch.zeros((1, 10)), torch.zeros((1, 3)))

        ax = None
        demo.display_hand(
            {"verts": verts.detach(), "joints": joints.detach()}, mano_faces=mano_layer.th_faces, ax=ax, show=False
        )
        plt.savefig("outputs/check/{}.jpg".format(img_path))
        # plt.cla()
        # plt.clf()
        plt.close()

    if label is not None:
        after_kpts = decode_keypoint(label)
    else:
        after_kpts = joint[:, :2]
    test_image = (img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
    test_image = cv2.cvtColor(test_image, cv2.COLOR_GRAY2RGB)
    test_image2 = test_image.copy()
    test_image = draw_skeleton(test_image, np.array(after_kpts) * 8)
    test_image2 = draw_skeleton(test_image2, np.array(joint[:, :2]) * 8)

    test_image = cv2.resize(test_image, (128, 128))
    test_image2 = cv2.resize(test_image2, (128, 128))
    saveimg = np.concatenate([test_image, test_image2], axis=1)

    if has_common_dof:
        figimg = cv2.imread("outputs/check/{}.jpg".format(img_path))
        figimg = cv2.resize(figimg, (256, 256))
        saveimg = np.concatenate([saveimg, figimg], axis=0)

    cv2.imwrite("outputs/check/{}.jpg".format(img_path), saveimg)


def check_dof_img_and_history_to_save(img_path, img, prev_img, joint, prev_joint, pose, prev_pose, has_common_dof):
    import torch
    from manopth import demo
    from manopth.manolayer import ManoLayer
    from matplotlib import pyplot as plt

    img_path = img_path.replace("/", "#")

    if has_common_dof:
        mano_layer = ManoLayer(
            mano_root="extra_data/manodatas/models",
            use_pca=False,
            flat_hand_mean=False,
        )

        pose = torch.tensor(pose.astype(np.float32))
        verts, joints3d = mano_layer.forward(pose.reshape(-1, 48), torch.zeros((1, 10)), torch.zeros((1, 3)))

        ax = None
        demo.display_hand(
            {"verts": verts.detach(), "joints": joints3d.detach()}, mano_faces=mano_layer.th_faces, ax=ax, show=False
        )
        plt.savefig("outputs/check/{}.jpg".format(img_path))

        prev_pose = torch.tensor(prev_pose.astype(np.float32))
        verts, joints3d = mano_layer.forward(prev_pose.reshape(-1, 48), torch.zeros((1, 10)), torch.zeros((1, 3)))

        ax = None
        demo.display_hand(
            {"verts": verts.detach(), "joints": joints3d.detach()}, mano_faces=mano_layer.th_faces, ax=ax, show=False
        )
        plt.savefig("outputs/check/{}.jpg".format("prev_" + img_path))

    test_image = (img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
    test_image = cv2.cvtColor(test_image, cv2.COLOR_GRAY2RGB)
    test_image = draw_skeleton(test_image, np.array(joint[:, :2]) * 8)
    test_image2 = (prev_img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
    test_image2 = cv2.cvtColor(test_image2, cv2.COLOR_GRAY2RGB)
    test_image2 = draw_skeleton(test_image2, np.array(prev_joint[:, :2]) * 8)

    test_image = cv2.resize(test_image, (128, 128))
    test_image2 = cv2.resize(test_image2, (128, 128))
    saveimg = np.concatenate([test_image, test_image2], axis=1)

    if has_common_dof:
        figimg = cv2.imread("outputs/check/{}.jpg".format(img_path))
        figimg = cv2.resize(figimg, (256, 256))
        prev_figimg = cv2.imread("outputs/check/{}.jpg".format("prev_" + img_path))
        prev_figimg = cv2.resize(prev_figimg, (256, 256))

        saveimg = np.concatenate([saveimg, figimg, prev_figimg], axis=0)

    cv2.imwrite("outputs/check/{}.jpg".format(img_path), saveimg)


def check_dof_img_and_history_to_save_multi_dataloader(transform_data_list):
    import torch
    from manopth import demo
    from manopth.manolayer import ManoLayer
    from matplotlib import pyplot as plt

    mano_layer = ManoLayer(
        mano_root="extra_data/manodatas/models",
        use_pca=False,
        flat_hand_mean=False,
    )

    img_path = ""
    # img_path = transform_data_list[0][0][1]["file_name"].replace("/", "_")
    big_imgs = []
    # big_pose_vis = []
    for samples_t in transform_data_list:
        img_views = []
        pose_vis_views = []
        for sample_view in samples_t:
            img, anno, targets = sample_view
            img = (img.add(0.5) * 255).cpu().data.numpy()[0].astype(np.uint8)
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
            joint2d = anno["joints_25d"].reshape(-1, 3)
            img = draw_skeleton(img, np.array(joint2d[:, :2]))
            img = cv2.resize(img, (128, 128))
            img_views.append(img)

            img_path += anno["file_name"].replace("/", "_")

            pose = anno["common_pose"]
            shape = anno["common_shape"]
            trans = anno["common_trans"]

            pose = torch.tensor(pose.astype(np.float32))
            shape = torch.tensor(shape.astype(np.float32))
            trans = torch.tensor(trans.astype(np.float32))
            verts, joints3d = mano_layer.forward(pose.reshape(-1, 48), shape.reshape(-1, 10), trans.reshape(-1, 3))

            Kvirt = anno["camI"]
            real_joint3d = anno["real_joint_3d"]
            joints2d_proj = np.matmul(real_joint3d, Kvirt.T)
            joints2d_proj /= joints2d_proj[:, 2:3]

            ax = None
            demo.display_hand(
                {"verts": verts.detach(), "joints": joints3d.detach()},
                mano_faces=mano_layer.th_faces,
                ax=ax,
                show=False,
            )
            plt.savefig("outputs/check/{}.jpg".format("tmp"))
            figimg = np.asarray(plt.gcf().canvas.buffer_rgba())[:, :, :3]
            plt.clf()
            # plt.savefig("outputs/check/{}.jpg".format(img_path))
            # figimg = cv2.imread("outputs/check/{}.jpg".format(img_path))
            figimg = cv2.resize(figimg, (256, 256))
            pose_vis_views.append(figimg)

        img_views = np.concatenate(img_views, axis=1)
        pose_vis_views = np.concatenate(pose_vis_views, axis=0)
        big_imgs.append(np.concatenate([img_views, pose_vis_views], axis=0))

    saveimg = np.concatenate(big_imgs, axis=0)
    tag = img_path.split("+")[0:2]
    tag = tag[0] + "+" + tag[1] + "+"
    cv2.imwrite("outputs/check/{}.jpg".format(tag + img_path.replace(tag, "").replace(".png", "")), saveimg)


def cal_iou(box1, box2):
    def box_area(box):
        return (box[2] - box[0]) * (box[3] - box[1])

    area1 = box_area(box1)
    area2 = box_area(box2)

    inter_x1 = max(box1[0], box2[0])
    inter_x2 = min(box1[2], box2[2])
    inter_y1 = max(box1[1], box2[1])
    inter_y2 = min(box1[3], box2[3])
    if inter_y2 - inter_y1 <= 0 or inter_x2 - inter_x1 <= 0:
        return 0
    inter_area = (inter_y2 - inter_y1) * (inter_x2 - inter_x1)
    return inter_area / (area1 + area2 - inter_area)  # iou = inter / (area1 + area2 - inter)


def filename_hash(filename):
    file_name_split = filename.split("/")[-3:]
    file_name = ""
    for x in file_name_split:
        file_name = file_name + "/" + x
    import hashlib

    file_name_hash = int(hashlib.sha1(file_name.encode("utf-8")).hexdigest(), 16) % (2**32)
    return file_name_hash


def extend_bbox(bbox, image_shape=None):
    x1, y1, x2, y2 = bbox
    wb, hb = x2 - x1, y2 - y1
    scale = 0.15

    if image_shape is not None:
        height, width = image_shape
        newx1 = x1 - wb * scale if x1 - wb * scale > 0 else 0
        newx2 = x2 + wb * scale if x2 + wb * scale < width else width - 1
        newy1 = y1 - hb * scale if y1 - hb * scale > 0 else 0
        newy2 = y2 + hb * scale if y2 + hb * scale < height else height - 1
    else:
        newx1 = x1 - wb * scale
        newx2 = x2 + wb * scale
        newy1 = y1 - hb * scale
        newy2 = y2 + hb * scale
    exbox = [int(newx1), int(newy1), int(newx2), int(newy2)]
    return exbox


def box_ioa(rect1, rect2):
    xmin1, ymin1, xmax1, ymax1 = rect1
    xmin2, ymin2, xmax2, ymax2 = rect2
    # s1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    s2 = (xmax2 - xmin2) * (ymax2 - ymin2)
    left = max(xmin2, xmin1)
    right = min(xmax2, xmax1)
    top = max(ymin2, ymin1)
    bottom = min(ymax2, ymax1)

    if left >= right or top >= bottom:
        return 0
    intersection = (right - left) * (bottom - top)
    return intersection / s2


def select_proposal(proposals, bbox, img_shape):
    candidate_box = list()
    candidate_box.append(extend_bbox(bbox, img_shape))
    for prop_conf in proposals:
        prop = prop_conf[:4]
        iou_pb = iou(prop, bbox)
        ioa_pb = box_ioa(prop, bbox)
        if iou_pb > 0.35 and ioa_pb > 0.8:
            candidate_box.append(prop)
    return candidate_box


def get_rectfy_rotation_angle(joint_2d):
    pt0 = joint_2d[0]
    pt1 = joint_2d[9]

    vec = pt1 - pt0
    angle = np.arctan2(vec[1], vec[0])

    if angle > 0:
        if angle - np.pi * 0.5 < 0:
            return 90, True
        else:
            return -90, False
    else:
        return 0, False


def rectfy_sample_by_rotate(img, joint_2d):
    height, width = img.shape[:2]

    angle, need_rot = get_rectfy_rotation_angle(joint_2d)

    if need_rot:
        if angle > 0:
            # roation 90 degree
            img = cv2.rotate(img, cv2.ROTATE_90_COUNTERCLOCKWISE)
            joint_2d_new = joint_2d.copy()
            joint_2d_new[:, 0] = joint_2d[:, 1]
            joint_2d_new[:, 1] = width - joint_2d[:, 0]
            # joint_2d_new_last = joint_2d_last.copy()
            # joint_2d_new_last[:, 0] = joint_2d_last[:, 1]
            # joint_2d_new_last[:, 1] = width_last - joint_2d_last[:, 0]
        else:
            img = cv2.rotate(img, cv2.ROTATE_90_CLOCKWISE)
            joint_2d_new = joint_2d.copy()
            joint_2d_new[:, 0] = height - joint_2d[:, 1]
            joint_2d_new[:, 1] = joint_2d[:, 0]
            # joint_2d_new_last = joint_2d_last.copy()
            # joint_2d_new_last[:, 0] = height_last - joint_2d_last[:, 1]
            # joint_2d_new_last[:, 1] = joint_2d_last[:, 0]

        joint_2d = joint_2d_new
        # joint_2d_last = joint_2d_new_last
        # joint_2d_last = joint_2d
    return img, joint_2d


def get_rot6s(pose):
    num_pose = pose.shape[0]
    rot6s = list()
    for i in range(num_pose):
        rot6 = cv2.Rodrigues(pose[i])[0][:2]
        rot6s.append(rot6.reshape(-1))
    rot6s = np.array(rot6s)
    return rot6s


def matrix2dof(matrixs):
    # batch_size = matrixs.shape[0]
    # dofs = list()
    # for i in range(batch_size):
    #     dof = cv2.Rodrigues(matrixs[i].numpy())
    #     dofs.append(dof[0])
    # dofs = np.array(dofs)
    dofs = matrix_to_axis_angle(matrixs).reshape(-1, 3, 1)
    return dofs


def rot6tomatrix(rot6s):
    poses = rot6s.reshape(-1, 6)
    x_raw = poses[:, 0:3]  # batch*3
    y_raw = poses[:, 3:6]  # batch*3

    x = normalize_vector(x_raw)  # batch*3
    y = normalize_vector(y_raw)  # batch*3
    middle = normalize_vector(x + y)
    orthmid = normalize_vector(x - y)
    x = normalize_vector(middle + orthmid)
    y = normalize_vector(middle - orthmid)
    # Their scalar product should be small !
    # assert torch.einsum("ij,ij->i", [x, y]).abs().max() < 0.00001
    z = normalize_vector(cross_product(x, y))

    x = x.view(-1, 3, 1)
    y = y.view(-1, 3, 1)
    z = z.view(-1, 3, 1)
    matrix = torch.cat((x, y, z), 2)  # batch*3*3
    # Check for reflection in matrix ! If found, flip last vector TODO

    # torch.det too slow, skip
    # assert (torch.stack([torch.det(mat) for mat in matrix]) < 0).sum() == 0

    return matrix


def normalize_vector(v):
    batch = v.shape[0]
    v_mag = torch.sqrt(v.pow(2).sum(1))  # batch
    v_mag = torch.max(v_mag, v.new([1e-8]))
    v_mag = v_mag.view(batch, 1).expand(batch, v.shape[1])
    v = v / v_mag
    return v


def cross_product(u, v):
    batch = u.shape[0]
    i = u[:, 1] * v[:, 2] - u[:, 2] * v[:, 1]
    j = u[:, 2] * v[:, 0] - u[:, 0] * v[:, 2]
    k = u[:, 0] * v[:, 1] - u[:, 1] * v[:, 0]

    out = torch.cat((i.view(batch, 1), j.view(batch, 1), k.view(batch, 1)), 1)

    return out


def rot6todof(rot6s):
    # rot6s = torch.from_numpy(rot6s)
    matrix = rot6tomatrix(rot6s)
    output_dofs = matrix2dof(matrix)
    output_dofs = output_dofs.reshape(-1, 3)

    return output_dofs * -1


def eular_to_mat(roll, pitch, yaw):
    rx = np.asarray(
        [
            [1.0, 0.0, 0.0],
            [0.0, np.cos(roll), -np.sin(roll)],
            [0.0, np.sin(roll), np.cos(roll)],
        ]
    )
    ry = np.asarray(
        [
            [np.cos(pitch), 0.0, np.sin(pitch)],
            [0.0, 1.0, 0.0],
            [-np.sin(pitch), 0.0, np.cos(pitch)],
        ]
    )
    rz = np.asarray(
        [
            [np.cos(yaw), -np.sin(yaw), 0.0],
            [np.sin(yaw), np.cos(yaw), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    return rx @ ry @ rz
